""" blah blah blah
"""

from clusterone.utils import render_table

def main(_data, header=None):
    """
    ???
    """
    #TODO: Pull the code from render_table here and refactor it, then test


    # Otherwise the table would broke
    assert len(header) == len(_data[0])

    table_data = [header] + [row for row in _data]

    table = render_table(table_data, 70).table
    return table




import os
import pytest
from click.testing import CliRunner
from click.exceptions import BadParameter

from clusterone import ClusteroneClient
from clusterone.persistance.session import Session
from clusterone.clusterone_cli import cli

from clusterone.commands.run.local import cmd
from clusterone.commands.run.local.cmd import validate_mode, validate_module, validate_env, validate_package_path


@pytest.mark.parametrize(
    'input, result', [
        ("new", False),
        ("current", True),
    ]
)
def test_validate_env(input, result):
    assert validate_env(None, None, input) == result


@pytest.mark.parametrize(
    'input, result', [
        ("single", "single-node"),
        ("distributed", "distributed"),
    ]
)
def test_validate_env(input, result):
    assert validate_mode(None, None, input) == result


@pytest.mark.parametrize(
    "input, result", [
        ("main.py", "main"),
        ("main", "main"),
        ("yelop", "yelop"),
    ]
)
def test_validate_module_strip_path_characters(input, result):
    assert validate_module(None, None, input) == result


@pytest.mark.parametrize(
    'input, result', [
        (".", ""),
        ("mnist", "mnist"),
        ("mnist/", "mnist"),
        ("something/mnist", "something.mnist"),
        ("folder/something/mnist", "folder.something.mnist"),
        ("even_more/folder/something/mnist", "even_more.folder.something.mnist"),
    ]
)
def test_validate_package_path(input, result):
    assert validate_package_path(None, None, input) == result


def test_command(mocker):

    def create_dummy_file(filename):
        with open(filename, "w") as file:
            file.write("pass\n")

    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    cmd.run_tf = mocker.Mock()
    runner = CliRunner()

    with runner.isolated_filesystem() as isolated_cwd:
        os.mkdir('./mnist')
        create_dummy_file("mnist/my_ml_experiment.py")
        create_dummy_file("ml-requirements.txt")

        result = runner.invoke(cli, [
            "run",
            "local",
            "single",
            "--module", "my_ml_experiment",
            "--worker-replicas", "8",
            "--ps-replicas", "3",
            "--requirements", "./ml-requirements.txt",
            "--env", "new",
            "--package-path", "mnist",
        ])

        # for debugging purposes
        print("Command output:\n", result.output)

        _, kwargs = cmd.run_tf.call_args

        assert kwargs == {
            "cwd": isolated_cwd,
            "mode": "single-node",
            "package_path": "mnist",
            "module": "my_ml_experiment",
            "worker_replicas": 8,
            "ps_replicas": 3,
            "requirements": "./ml-requirements.txt",
            "current_env": False,
        }



from collections import OrderedDict
import pytest
from click.exceptions import BadParameter

from clusterone import ClusteroneClient
from clusterone.clusterone_cli import Context
from clusterone.client_exceptions import JobNameConflict, NonExistantJob
from clusterone.persistance.session import Session

from .path_to import path_to_job_id
from .path_to import path_to_project
from .path_to import path_to_dataset

def test_user_not_specified(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    ClusteroneClient.get_project = mocker.Mock(return_value={"id": ""})
    ClusteroneClient.get_jobs = mocker.Mock(return_value=[OrderedDict([('job_id', '6789-6789-6789')])])
    mocker.patch.object(Session, 'get', autospec=True, return_value="someuser")
    client = ClusteroneClient()
    context = Context(client, Session(), None)

    path_to_job_id("my-tport-project/master", context)

    ClusteroneClient.get_project.assert_called_with("my-tport-project", "someuser")

def test_client_name_conflict(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    ClusteroneClient.get_project = mocker.Mock(return_value={"id": ""})
    ClusteroneClient.get_jobs = mocker.Mock(return_value=[OrderedDict([('job_id', '6789-6789-6789')]), OrderedDict([('job_id', '4567-4567-4567')])])
    mocker.patch.object(Session, 'get', autospec=True, return_value="someuser")
    client = ClusteroneClient()
    context = Context(client, Session(), None)

    with pytest.raises(JobNameConflict) as exception:
        path_to_job_id("allgreed/my-tport-project/job", context)

        # Warning, this depends on the JobNameConflict implementation, see id field
        assert exception.ids == ['6789-6789-6789', '4567-4567-4567']

def test_job_not_exist(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    ClusteroneClient.start_job = mocker.Mock()
    ClusteroneClient.get_project = mocker.Mock(return_value={"id": ""})
    ClusteroneClient.get_jobs = mocker.Mock(return_value=[])
    mocker.patch.object(Session, 'get', autospec=True, return_value="someuser")
    client = ClusteroneClient()
    context = Context(client, Session(), None)

    with pytest.raises(NonExistantJob):
        path_to_job_id("allgreed/my-tport-project/very-not-existant-job-path", context)

def test_resolve_job_id(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    ClusteroneClient.start_job = mocker.Mock()
    ClusteroneClient.get_project = mocker.Mock(return_value={"id": "12345"})
    ClusteroneClient.get_jobs = mocker.Mock(return_value=[OrderedDict([('job_id', '6789-6789-6789')])])
    mocker.patch.object(Session, 'get', autospec=True, return_value="someuser")
    client = ClusteroneClient()
    context = Context(client, Session(), None)

    path_to_job_id("allgreed/my-tport-project/master", context)

    ClusteroneClient.get_jobs.assert_called_with({
        "repository": "12345",
        "display_name": "master"
        })

    ClusteroneClient.get_project.assert_called_with("my-tport-project", "allgreed")

def test_invalid_job_path(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'get', autospec=True, return_value="someuser")
    client = ClusteroneClient()
    context = Context(client, Session(), None)

    with pytest.raises(BadParameter):
        path_to_job_id("elorapmordeczko//////xd", context)

"""
Test path_to_project
"""

def test_basic_project_path(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'get', autospec=True, return_value="someuser")
    client = ClusteroneClient()
    context = Context(client, Session(), None)
    ClusteroneClient.get_project = mocker.Mock(return_value={"id": "12345"})

    assert path_to_project("username/project_name", context) == {"id": "12345"}
    ClusteroneClient.get_project.assert_called_with("project_name", "username")

def test_invalid_path(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'get', autospec=True, return_value="someuser")
    client = ClusteroneClient()
    context = Context(client, Session(), None)
    ClusteroneClient.get_project = mocker.Mock(return_value={"id": "12345"})

    with pytest.raises(BadParameter):
        path_to_project("elorapmordeczko//////xd", context)

def test_default_username(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'get', autospec=True, return_value="someuser")
    client = ClusteroneClient()
    context = Context(client, Session(), None)
    ClusteroneClient.get_project = mocker.Mock(return_value={"id": "12345"})

    assert path_to_project("project_name", context) == {"id": "12345"}
    ClusteroneClient.get_project.assert_called_with("project_name", "someuser")

"""
Test path_to_dataset
"""

def test_basic_dataset_path_d(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'get', autospec=True, return_value="someuser")
    client = ClusteroneClient()
    context = Context(client, Session(), None)
    ClusteroneClient.get_dataset = mocker.Mock(return_value={"id": "12345"})

    assert path_to_dataset("username/dataset_name", context) == {"id": "12345"}
    ClusteroneClient.get_dataset.assert_called_with("dataset_name", "username")

def test_invalid_path_d(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'get', autospec=True, return_value="someuser")
    client = ClusteroneClient()
    context = Context(client, Session(), None)
    ClusteroneClient.get_dataset = mocker.Mock(return_value={"id": "12345"})

    with pytest.raises(BadParameter):
        path_to_dataset("elorapmordeczko//////xd", context)

def test_default_username_d(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'get', autospec=True, return_value="someuser")
    client = ClusteroneClient()
    context = Context(client, Session(), None)
    ClusteroneClient.get_dataset = mocker.Mock(return_value={"id": "12345"})

    assert path_to_dataset("dataset_name", context) == {"id": "12345"}
    ClusteroneClient.get_dataset.assert_called_with("dataset_name", "someuser")



from click.testing import CliRunner

from clusterone import ClusteroneClient
from clusterone.clusterone_cli import cli
from clusterone.just_client import Notebook


def test_command(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mock_notebook = mocker.Mock()
    mock_notebook.id = "853b9f10-36ce-4de4-b2f8-108d69733b42"
    mocker.patch.object(Notebook, 'from_clusterone', return_value=mock_notebook)

    result = CliRunner().invoke(cli, ['stop', 'notebook', 'dummy'])

    Notebook.from_clusterone.assert_called_with(mocker.ANY, "dummy")
    assert mock_notebook.stop.called
    assert result.output == "853b9f10-36ce-4de4-b2f8-108d69733b42\n"



from collections import OrderedDict

from click.testing import CliRunner

from clusterone import ClusteroneClient
from clusterone.clusterone_cli import cli
from clusterone.commands.rm.project import cmd
from clusterone.client_exceptions import BusyProjectRemoveAttempt

def test_deletion(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    ClusteroneClient.delete_project = mocker.Mock()
    ClusteroneClient.get_jobs = mocker.Mock(return_value=[])
    cmd.path_to_project = mocker.Mock(return_value=OrderedDict([
        ('id', 'some-dummy-id'),
        ('username', 'someDummyUsername'),
        ('name', 'some_dummy_project_name'),
    ]))
    mocker.spy(cmd, 'confirm')

    CliRunner().invoke(cli, ['rm', 'project', 'someDummyUsername/some_dummy_project_name'], 'y')

    cmd.path_to_project.assert_called_with(
        'someDummyUsername/some_dummy_project_name',
        context=mocker.ANY)

    cmd.confirm.assert_called_with(mocker.ANY, abort=True)

    ClusteroneClient.delete_project.assert_called_with(
        'some_dummy_project_name',
        'someDummyUsername')

def test_running_jobs(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    ClusteroneClient.delete_project = mocker.Mock()
    ClusteroneClient.get_jobs = mocker.Mock(return_value=[OrderedDict([('status', 'started')])])
    cmd.path_to_project = mocker.Mock(return_value=OrderedDict([
        ('id', 'some-dummy-id'),
        ('username', 'someDummyUsername'),
        ('name', 'some_dummy_project_name'),
    ]))
    mocker.spy(cmd, 'confirm')

    result = CliRunner().invoke(cli, ['rm', 'project', 'project'])

    assert isinstance(result.exception, BusyProjectRemoveAttempt)



from uuid import UUID
from coreapi.exceptions import ErrorMessage
from collections import namedtuple

from clusterone.just_client.types import TaskStatus
from clusterone.client_exceptions import _NonExistantNotebook, NonExistantNotebook, DuplicateNotebookName


NotebookPath = namedtuple("NotebookPath", "notebook_name username")


class Notebook(object):
    """
    Handle of jupyter notebook
    """

    def _initialize(self, api_response):
        """
        This works as __init__

        For production code make sure that the passed data is real, up to date response from Clusterone API
        """

        self.url = api_response['notebook_url']
        self.id = api_response['job_id']
        self.status = TaskStatus(api_response['status'])
        self.token = api_response['parameters']['token']

    def __init__(self, client, configuration):
        """
        Create a brand new notebook and return a handle

        The method is overwritten to have nice API, see _initialize for initialisation
        """

        params = {
                'repository': configuration['code'],
                'display_name': configuration['name'],
                'description': configuration['description'],
                'parameters': configuration,
                'datasets_set': configuration['datasets_set'],
            }

        try:
            params['git_commit_hash'] = configuration['git_commit_hash']
        except KeyError:
            pass

        try:
            response = client.client_action(['notebooks', 'create'], params=params)
        except ErrorMessage as exception:
            if "400" in exception.error.title:
                raise DuplicateNotebookName()
            else:
                raise

        self._initialize(response)

    @classmethod
    def _from_data(cls, api_response):
        """
        Alternate constructor that bypasses calling __init__ and uses _initialize instead

        Also: might be useful for testing.
        """

        # to avoid triggering __init__
        bare_instance = cls.__new__(cls)

        bare_instance._initialize(api_response)
        return bare_instance

    @classmethod
    def from_clusterone(cls, client, path_or_uuid):
        """
        Obtains the handle of existing jupyter notebook from Clusterone.

        The handle can be obtained by providing either uuid or path.
        """

        try:
            UUID(path_or_uuid)
        except ValueError:
            notebook_path = cls._parse_notebook_path(path_or_uuid)

            request_target = "notebook_by_name"
            request_params = {
                "display_name": notebook_path.notebook_name,
                # if explicit username is not provided use the currently logged user
                "username": notebook_path.username or client.username,
            }
            not_found_exception = NonExistantNotebook
        else:
            request_target = "notebooks"
            request_params = {
                'job_id': path_or_uuid,
            }
            not_found_exception = _NonExistantNotebook

        try:
            response = client.client_action([request_target, 'read'], params=request_params)
            # no need to check for duplicates, the notebook name is guaranteed to be unique by the backend
        except ErrorMessage as exception:
            if "404" in exception.error.title:
                raise not_found_exception()
            else:
                raise

        return cls._from_data(response)

    @staticmethod
    def _parse_notebook_path(path_string):
        """
        Extracts notebook name and optionally username from user provided string
        """

        parsed_path_components = path_string.split('/')

        if len(parsed_path_components) > 2:
            raise ValueError("Path string not conforming to the 'username/notebook_name' or 'notebook_name' format")

        if not parsed_path_components[0]:
            raise ValueError("Notebook name cannot be empty")

        try:
            return NotebookPath(username=parsed_path_components[0], notebook_name=parsed_path_components[1])
        except IndexError:
            return NotebookPath(username=None, notebook_name=parsed_path_components[0])

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return "Notebook(<client_instance>, {!r})".format(self.id)

    def start(self, client):
        client.client_action(['notebooks', 'start'], params={
            'job_id': self.id,
        })

        self.status = TaskStatus.started

    def stop(self, client):
        client.client_action(['notebooks', 'stop'], params={
            'job_id': self.id,
        })

        self.status = TaskStatus.stopped



from click.testing import CliRunner

from clusterone import ClusteroneClient
from clusterone.clusterone_cli import cli
from clusterone.commands.config import cmd
from clusterone.persistance.config import Config
from clusterone.persistance.session import Session


def test_passing(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'save', autospec=True, return_value=None)
    mocker.patch.object(Config, '__init__', autospec=True, return_value=None)
    cmd.setattr = mocker.Mock()

    CliRunner().invoke(cli, [
        'config',
        'endpoint',
        'dummy-value',
    ])

    args, kwargs = cmd.setattr.call_args
    assert isinstance(args[0], Config)
    assert args[1] == 'endpoint'
    assert args[2] == 'dummy-value'


def test_non_configurable(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'save', autospec=True, return_value=None)
    mocker.patch.object(Config, '__init__', autospec=True, return_value=None)
    cmd.setattr = mocker.Mock()

    result = CliRunner().invoke(cli, [
        'config',
        'dummy-key',
        'dummy-value',
    ])

    assert result.exit_code == 2



import pytest
from .time_limit_to_minutes import main

def test_hours_only():
    assert main("5h") == 300

def test_minutes_only():
    assert main("22m") == 22

def test_hours_and_minutes():
    assert main("5h22m") == 322

def test_different_case():
    assert main("5H22m") == 322
    assert main("5h22M") == 322
    assert main("5H22M") == 322

def test_invalid_path():
    with pytest.raises(ValueError):
        main('xxx22h55mxxxxx')



import sys

import click
from click.exceptions import MissingParameter, BadParameter

from clusterone import authenticate
from clusterone.utilities import silent_prompt
from .helper import main


def is_data_on_stdin():
    return not sys.stdin.isatty()

@click.command()
@click.option(
    '--project-path',
    '-p',
    help="Project identifier. Format \"username/project\".")
@click.option(
    '--repo-path',
    '-r',
    help="Path to local Git repository.")
@click.pass_obj
@authenticate()
def command(context, project_path, repo_path):
    """
    Links existing Clusterone project with a local Git repository
        Clusterone remote url can be passed through stdin.
    """

    client, session, cwd = context.client, context.session, context.cwd

    # since context cannot be referenced in the deault paramters
    repository_path = repo_path if repo_path is not None else cwd

    if is_data_on_stdin():
        remote_url = silent_prompt()
    else:
        if project_path is None:
            raise MissingParameter(
                "Please provide either of them.",
                param_type="remote link or project-path")

        path_tokens = project_path.split('/')

        if not len(path_tokens) == 2:
            if len(path_tokens) == 1:
                path_tokens = [session.get("username")] + path_tokens
            else:
                raise BadParameter(param_hint="--project-path", message="Please provide a valid project-path. Format: \"username/project\".")

        username, project_name = path_tokens

        project = client.get_project(project_name, username=username)

        remote_url = project.get('git_auth_link')

    main(context, repository_path, remote_url)
    return



from collections import OrderedDict

import pytest
from click.testing import CliRunner

from clusterone import ClusteroneClient
from clusterone.clusterone_cli import cli
from clusterone.commands.get.job import cmd


def test_job_aquisition(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    ClusteroneClient.get_job = mocker.Mock()
    cmd.serialize_job = mocker.Mock(return_value="12345678")

    CliRunner().invoke(cli, ['get', 'job', 'project/job_name'])

    cmd.serialize_job.assert_called_with('project/job_name', context=mocker.ANY)

def test_prop_stripping_single(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    cmd.serialize_job = mocker.Mock(return_value=OrderedDict([
    ('job_id', 'e144f11c-0207-47d6-b672-7efd692b0885'),
    ('repository', 'aaf4de71-f506-48c0-855c-02c7c485c5a4'),
    ('repository_owner', 'allgreed'),
    ('repository_name', 'mnist-demo'),
    ('repository_owner_photo_url',
     'https://clusterone-api-development.s3.amazonaws.com/images/empty_image_400_400.png'
     ),
    ('display_name', 'job'),
    ('description', 'Dummy job'),
    ('created_at', '2018-01-12T12:38:23.593018Z'),
    ('launched_at', '2018-01-24T16:33:19.411485Z'),
    ('terminated_at', '2018-01-24T16:41:41.557648Z'),
    ('modified_at', '2018-01-24T16:43:32.631078Z'),
    ('git_commit_hash', '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1'),
    ('status', 'stopped'),
    ('parameters', OrderedDict([
        ('debug', ''),
        ('package_manager', 'pip'),
        ('worker_type', 't2.small'),
        ('workers', OrderedDict([
            ('slug', 't2.small'),
            ('replicas', 1),
            ('queue', 'tensorport-jobmaster'),
            ('type', 't2.small'),
            ('blessed', False)
        ])),
        ('datasets', []),
        ('code_repo', OrderedDict([('hash',
         '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1'), ('url',
         'https://git.clusterone.com/allgreed/mnist-demo.git'),
         ('mount_point', 'code')])),
        ('python_version', '2.7'),
        ('package_path', ''),
        ('tf_version', '1.0.0'),
        ('requirements', 'requirements.txt'),
        ('code_repo_commit', '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1'
         ),
        ('time_limit', 2880),
        ('git_username', 'allgreed'),
        ('git_password', '584ed7f63b75c50ef3d38e9514d2cd14dc39eb14'),
        ('module', 'main'),
        ('worker_replicas', 2),
        ('framework', OrderedDict([
            ('slug', 'tensorflow-1.3.0'),
            ('name', 'tensorflow'),
            ('version', '1.3.0'),
        ])),
        ('mode', 'single'),
        ])),
    ]))

    result = CliRunner().invoke(cli, ['get', 'job', 'project/job_name'])

    assert result.output != ''
    assert not any([(parameter in result.output) for parameter in ["Ps type", "Worker replicas", "Ps replicas"]])


@pytest.mark.skip("Hotfixing")
def test_table_call(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    cmd.serialize_job = mocker.Mock(return_value=OrderedDict([
    ('job_id', 'e144f11c-0207-47d6-b672-7efd692b0885'),
    ('repository', 'aaf4de71-f506-48c0-855c-02c7c485c5a4'),
    ('repository_owner', 'allgreed'),
    ('repository_name', 'mnist-demo'),
    ('repository_owner_photo_url',
     'https://clusterone-api-development.s3.amazonaws.com/images/empty_image_400_400.png'
     ),
    ('display_name', 'job'),
    ('description', 'Dummy job'),
    ('created_at', '2018-01-12T12:38:23.593018Z'),
    ('launched_at', '2018-01-24T16:33:19.411485Z'),
    ('terminated_at', '2018-01-24T16:41:41.557648Z'),
    ('modified_at', '2018-01-24T16:43:32.631078Z'),
    ('git_commit_hash', '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1'),
    ('status', 'stopped'),
    ('parameters', OrderedDict([
        ('debug', ''),
        ('package_manager', 'pip'),
        ('worker_type', 't2.small'),
        ('workers', OrderedDict([
            ('slug', 't2.small'),
            ('replicas', 2),
            ('queue', 'tensorport-jobmaster'),
            ('type', 't2.small'),
            ('blessed', False)
        ])),
        ('parameter_servers', OrderedDict([
            ('slug', 't2.small'),
            ('replicas', 1),
            ('queue', 'tensorport-jobmaster'),
            ('type', 't2.small'),
            ('blessed', False)
        ])),
        ('datasets', []),
        ('code_repo', OrderedDict([('hash',
         '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1'), ('url',
         'https://git.clusterone.com/allgreed/mnist-demo.git'),
         ('mount_point', 'code')])),
        ('python_version', '2.7'),
        ('package_path', ''),
        ('tf_version', '1.0.0'),
        ('requirements', 'requirements.txt'),
        ('code_repo_commit', '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1'
         ),
        ('time_limit', 2880),
        ('git_username', 'allgreed'),
        ('git_password', '584ed7f63b75c50ef3d38e9514d2cd14dc39eb14'),
        ('module', 'main'),
        ('worker_replicas', 2),
        ('framework', OrderedDict([
            ('slug', 'tensorflow-1.3.0'),
            ('name', 'tensorflow'),
            ('version', '1.3.0'),
        ])),
        ('mode', 'distributed'),
        ])),
    ]))
    cmd.path_to_job_id = mocker.Mock(return_value="12345678")
    cmd.make_table = mocker.Mock(return_value="Example table string")

    CliRunner().invoke(cli, ['get', 'job', 'project/job_name'])

    cmd.make_table.assert_called_with([('Name', 'job'), ('Status', 'stopped'), ('Project', 'mnist-demo'), ('Module', 'main'), ('Package-path', ''),('Datasets', ''), ('Python version', '2.7'), ('Package manager', 'pip'), ('Requirements', 'requirements.txt'), ('Framework', 'tensorflow'), ('Framework version', '1.3.0'), ('Mode', 'distributed'), ('Worker type', 't2.small'),('Worker replicas', 2),('Ps type', 't2.small'),('Ps replicas', 1), ('Time limit', '2880 minutes')], header=['Property', 'Value'])

def test_printing_out_table(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    cmd.serialize_job = mocker.Mock(return_value=OrderedDict([
    ('job_id', 'e144f11c-0207-47d6-b672-7efd692b0885'),
    ('repository', 'aaf4de71-f506-48c0-855c-02c7c485c5a4'),
    ('repository_owner', 'allgreed'),
    ('repository_name', 'mnist-demo'),
    ('repository_owner_photo_url',
     'https://clusterone-api-development.s3.amazonaws.com/images/empty_image_400_400.png'
     ),
    ('display_name', 'job'),
    ('description', 'Dummy job'),
    ('created_at', '2018-01-12T12:38:23.593018Z'),
    ('launched_at', '2018-01-24T16:33:19.411485Z'),
    ('terminated_at', '2018-01-24T16:41:41.557648Z'),
    ('modified_at', '2018-01-24T16:43:32.631078Z'),
    ('git_commit_hash', '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1'),
    ('status', 'stopped'),
    ('parameters', OrderedDict([
        ('debug', ''),
        ('package_manager', 'pip'),
        ('worker_type', 't2.small'),
        ('workers', OrderedDict([
            ('slug', 't2.small'),
            ('replicas', 2),
            ('queue', 'tensorport-jobmaster'),
            ('type', 't2.small'),
            ('blessed', False)
        ])),
        ('parameter_servers', OrderedDict([
            ('slug', 't2.small'),
            ('replicas', 1),
            ('queue', 'tensorport-jobmaster'),
            ('type', 't2.small'),
            ('blessed', False)
        ])),
        ('datasets', []),
        ('code_repo', OrderedDict([('hash',
         '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1'), ('url',
         'https://git.clusterone.com/allgreed/mnist-demo.git'),
         ('mount_point', 'code')])),
        ('python_version', '2.7'),
        ('package_path', ''),
        ('tf_version', '1.0.0'),
        ('requirements', 'requirements.txt'),
        ('code_repo_commit', '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1'
         ),
        ('time_limit', 2880),
        ('git_username', 'allgreed'),
        ('git_password', '584ed7f63b75c50ef3d38e9514d2cd14dc39eb14'),
        ('module', 'main'),
        ('worker_replicas', 2),
        ('framework', OrderedDict([
            ('slug', 'tensorflow-1.3.0'),
            ('name', 'tensorflow'),
            ('version', '1.3.0'),
        ])),
        ('mode', 'distributed'),
        ])),
    ]))
    cmd.path_to_job_id = mocker.Mock(return_value="12345678")
    cmd.make_table = mocker.Mock(return_value="Example table string")

    result = CliRunner().invoke(cli, ['get', 'job', 'project/job_name'])

    assert result.output == "Example table string\n"



from . import main
silent_prompt = main.silent_prompt

def test_silence(capsys, monkeypatch):
    " Does not output to stdout "

    monkeypatch.setattr('sys.stdin.readline', lambda: "Mark")
    silent_prompt()

    captured = capsys.readouterr()

    assert captured.out == ""
    assert captured.err == ""

def test_calls_prompt(mocker):
    main.prompt = mocker.Mock()

    silent_prompt()

    assert main.prompt.called

def test_calls_with_aditional_args(mocker):
    main.prompt = mocker.Mock()

    silent_prompt(testparam="Test value")

    args, kwargs = main.prompt.call_args
    assert kwargs['testparam'] == "Test value"

# Test random name generator

def test_random_job_name(mocker):
    random_name = main.random_job_name()

    tokens = random_name.split('-')
    assert tokens[0] in main.adjs
    assert tokens[1] in main.nouns
    # Range excludes the last number, so 999 would fails, though it's ok
    assert tokens[2] in [str(_) for _ in range(main.MIN_NUM, main.MAX_NUM+1)]



import click

from clusterone import authenticate

from clusterone.messages import project_creation_in_progress_message
from .helper import main


@click.command()
@click.argument('name')
@click.option('--description', default='')
@click.pass_obj
@authenticate()
def command(context, name, description):
    """
    Create a new Clusterone project and output its git remote URL
    """

    click.echo(project_creation_in_progress_message)

    remote_url = main(context, name, description)

    click.echo(remote_url)

    return remote_url




from collections import namedtuple
from enum import Enum


class Output(Enum):
    NONE = 0
    ANY = 1
    CHECKPOINT = 2
    LOSS = 3


UserCredentials = namedtuple("UserCredentials", "username password email")



import click
from click import echo

from clusterone.utilities import Choice
from clusterone import client
from clusterone.just_client import Notebook
from clusterone.commands.create.job.base_cmd import job_base_options, base


@job_base_options(project_required=False)
@click.option(
    '--instance-type',
    type=Choice(client.instance_types_slugs),
    default="t2.small",
    help="Type of single instance to run.")
def command(context, custom_arguments, **kwargs):
    """
    Create a Jupyter notebook.
    """

    notebook_configuration = base(context, kwargs, module_arguments=custom_arguments)

    notebook_configuration['parameters']['mode'] = "single"
    notebook_configuration['parameters']['workers'] = \
        {
            'slug': kwargs['instance_type'],
            'replicas': 1
        }

    notebook_configuration['parameters']['framework'] = {'slug': 'jupyter'}

    # flattens notebook_configuration
    _notebook_configuration = notebook_configuration['parameters']
    _notebook_configuration.update(notebook_configuration['meta'])
    notebook_configuration = _notebook_configuration

    notebook = Notebook(client, notebook_configuration)
    echo(notebook.url)
    echo("access token: {}".format(notebook.token))



import time

from clusterone.client_exceptions import RemoteAquisitionFailure


def aquire_remote(project_name, session, client):

    username = session.get('username')
    retry_count, retry_interval = session.retry_count, session.retry_interval

    for _ in range(retry_count):
        time.sleep(retry_interval)

        project = client.get_project(project_name, username)
        git_url = project.get('git_auth_link')

        # API response might be evaluated as None or ""
        if not (git_url is None or git_url == ""):
            return git_url

    raise RemoteAquisitionFailure()

def main(context, name, description):

    client, session = context.client, context.session

    project_name = client.create_project(name, description)
    project_url = aquire_remote(project_name, client=client, session=session)
    return project_url



from contextlib import suppress

from clusterone import ClusteroneClient
from .actions import register_user, login, init_gitlab_project_from_clusterone_samples, start_job, create_job, set_endpoint, logout, \
    check_outputs
from .scenario import Scenario
from .types import Output
from .utilities import generate_random_credentials, authorize_client_via_session_file, poll


class GettingStarted(Scenario):
    """
    Mimicking programmatically our onboarding tutorial
    """

    def __init__(self):
        self.url = "https://clusterone.com/api/"

        self.project_name = "clusterone-test-onboarding"
        self.project_url = "https://github.com/clusterone/self-driving-demo"

        self.job_name = "clusterone-test-job"

        self.testing_credentials = generate_random_credentials()

        self.any_output_timeout = 600  # seconds
        self.checkpoint_output_timeout = 600  # seconds
        self.loss_output_timeout = 600  # seconds
        self.poll_interval = 10  # seconds

        self.client = ClusteroneClient(username=self.testing_credentials.username, api_url=self.url)

        logout()
        set_endpoint(self.url)

    def run(self):
        register_user(self.testing_credentials, client=self.client)
        login(self.testing_credentials)

        # This is a hax
        authorize_client_via_session_file(client=self.client)

        init_gitlab_project_from_clusterone_samples(name=self.project_name, project_url=self.project_url)

        create_job(self.project_name, self.job_name, module="main_tf", framework="tensorflow-1.3.0", dataset="tensorbot/self-driving-demo-data")
        start_job(self.project_name, self.job_name)

        # TODO: Test for events to know when the pods have started
        # TODO: After that reduce the timeout on any outputs

        poll(check_outputs, args=(Output.ANY, self.job_name), kwargs={"client": self.client}, timeout=self.any_output_timeout,
            interval=self.poll_interval)

        poll(check_outputs, args=(Output.CHECKPOINT, self.job_name), kwargs={"client": self.client},
             timeout=self.checkpoint_output_timeout,
             interval=self.poll_interval)

        poll(check_outputs, args=(Output.LOSS, self.job_name), kwargs={"client": self.client},
             timeout=self.loss_output_timeout,
             interval=self.poll_interval)

    def clean(self):
        # TODO: Do this better
        from .utilities import just_exec, call

        with suppress(AssertionError):
            # TODO: move this to actions
            just_exec("stop job -p {}/{}".format(self.project_name, self.job_name))

        # TODO: move this to actions
        call("rm -rf self-driving-demo", asserted=False)



FRAMEWORK_SLUGS = lambda: ['pytorch-1.0.0', 'tensorflow-1.4.0', 'tensorflow-1.5.0', 'tensorflow-1.3.0', 'tensorflow-1.2.0', 'tensorflow-1.0.0']
INSTANCE_SLUGS = lambda: ['c4.2xlarge', 'c4.2xlarge-spot', 'p2.xlarge', 'p2.xlarge-spot', 'p3.2xlarge', 't2.small']
PS_SLUGS = lambda: ['c4.2xlarge', 'c4.2xlarge-spot', 'p2.xlarge-spot', 't2.small']




from click.testing import CliRunner

from clusterone import ClusteroneClient
from clusterone.persistance.session import Session
from clusterone.clusterone_cli import cli
from clusterone.client_exceptions import NotSupported

from clusterone.commands.login import cmd


def test_upgrade_required(mocker):
    """
    PyPi version is newer than current CLI version
    """

    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'save', autospec=True, return_value=None)
    ClusteroneClient.api_login = mocker.Mock()
    ClusteroneClient.token = "sample_token"
    ClusteroneClient.username = "username"
    cmd.echo = mocker.Mock()

    cmd.is_latest_version = mocker.Mock(return_value=False)

    result = CliRunner().invoke(cli, ['login'], 'username\npassword',)
    assert isinstance(result.exception, NotSupported)

def test_no_upgrade(mocker):
    """
    PyPi version is the same or older than current CLI version
    """

    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'save', autospec=True, return_value=None)
    ClusteroneClient.api_login = mocker.Mock()
    ClusteroneClient.token = "sample_token"
    ClusteroneClient.username = "username"
    cmd.echo = mocker.Mock()

    cmd.is_latest_version = mocker.Mock(return_value=True)
    cmd.perform_version_check = mocker.Mock()

    result = CliRunner().invoke(cli, ['login'], 'username\npassword',)

    assert cmd.perform_version_check.called
    assert result.exception is None

def test_calls(mocker):
    """
    Dependency calls
    """

    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'save', autospec=True)
    ClusteroneClient.api_login = mocker.Mock()
    ClusteroneClient.token = "sample_token"
    ClusteroneClient.username = "username"
    cmd.echo = mocker.Mock()

    cmd.is_latest_version = mocker.Mock(return_value=True)

    CliRunner().invoke(cli, ['login'], 'username\npassword',)

    ClusteroneClient.api_login.assert_called_with("username", "password")
    assert Session.save.called
    cmd.echo.assert_called_with("Login successful")



import re as regexp
from collections import OrderedDict
from functools import reduce

import click
from click.exceptions import BadParameter, MissingParameter

from clusterone import client, authenticate
from clusterone.utilities import random_job_name, path_to_project, time_limit_to_minutes, path_to_dataset, Choice, \
    append_to_docstring

PYTHON_VERSION_DEFAULTS = {
    "2": "2.7",
    "3": "3.6",
}


# TODO: Move this to custom types
def validate_name(context, parameters, value):
    if not regexp.match("^[a-zA-Z0-9_-]+$", value):
        raise BadParameter("Should only contain alphanumeric characters, \"_\", or \"-\".")

    return value


def validate_time_limit(context, parameters, value):
    try:
       return time_limit_to_minutes(value)
    except ValueError:
        raise BadParameter("Please conform to [hours]h[minutes]m format, e.g. \"20h12m\".")


def validate_module_arguments(context, parameters, raw_value):
    arguments = [argument.lstrip('-') for argument in raw_value[::2]]
    values = raw_value[1::2]

    if len(arguments) != len(values):
        raise MissingParameter("", param_type="one or more values for custom argument")

    return {argument: value for argument, value in zip(arguments, values)}


def combine_options(options):
    """
    Generates a single decorator out of list of options decorator
    """

    def wrapper(function):
        return reduce(lambda decoratee, option_decorator: option_decorator(decoratee), reversed(options), function)
    return wrapper


_job_base_options = [
    # This is used to process user defined arguments
    # See to learn more: http://click.pocoo.org/6/advanced/#forwarding-unknown-options
    click.command(context_settings=dict(
        ignore_unknown_options=True,
    )),
    click.argument('custom-arguments', nargs=-1, callback=validate_module_arguments),
    # Documenting the above behavior to the user
    append_to_docstring('\n\nCustom arguments can be passed via "--myarg myvalue" syntax.'),

    # Common options for implementation
    click.pass_obj,
    authenticate(),

    # Business logic common options
    click.option(
        '--name',
        default=random_job_name(),
        callback=validate_name,
        help='Name of the job to be created',
        ),
    click.option(
        '--commit',
        help="Hash of commit to be run. Default: latest",
        ),
    click.option(
        '--datasets',
        help="Comma-separated list of the datasets to use for the job. Format: \"username/dataset-name\", e.g. 'clusterone/mnist-training:[GIT COMMIT HASH],clusterone/mnist-val:[GIT COMMIT HASH]'",
        default="",
        ),
    click.option(
        '--module',
        default='main',
        help='Name of module to run. Default: \"main\"',
        ),
    click.option(
        '--package-path',
        # The project root is represented by API as None
        help='Path to module within the project. Default: project root',
        ),
    click.option(
        '--python-version',
        type=Choice(['2', '2.7', '3', '3.5', '3.6']),
        default='2.7',
        help='Python version to use',
        ),
    click.option(
        '--framework',
        type=Choice(client.framework_slugs),
        default="tensorflow-1.3.0",
        help='Framework to be used. Default: tensorflow',
        ),
    #TODO: test passing
    #TODO: test default
    click.option(
        '--package-manager',
        type=Choice(['pip', 'conda', 'anaconda']),
        default="pip",
        help='Package manager to use. Default: pip',
        ),
    click.option(
        '--requirements',
        help="Requirements file to use. Default: requirements.txt",
        ),
    #TODO: Create custom click type time
    # https://github.com/click-contrib/click-datetime ?
    click.option(
        '--time-limit',
        default="48h",
        callback=validate_time_limit,
        help="Time limit for the job. Format: [hours]h[minutes]m, e.g. \"22h30m\""
        ),
    click.option(
        '--description',
        default="",
        help='Job description'
        ),
    click.option(
        '--gpu-count',
        type=click.IntRange(0, float('inf')),
        help="Number of instance GPUs to be used. Defaults to max instance's GPU count",
        ),
    ]


def job_base_options(project_required=True):
    """
    Provides customization on the base
    """

    project_option = click.option(
        '--project',
        'project_path',
        required=project_required,
        help="Project path to be run. Format: \"username/project-name\"",
    )

    return combine_options(_job_base_options + [project_option])


    #TODO: Python2.7 compliance legacy - add * after last positional argument
def base(context, kwargs, module_arguments=None):
    """
    Common option processing login for all job types
    """

    # TODO: remove this from client AND move the functionality to the API
    if kwargs['project_path']:
        project = path_to_project(kwargs['project_path'], context=context)
        project_id = project['id']
    else:
        project_id = None

    #TODO: Refactor this like hell!
    #TODO: OMG TEST THIS
    #TODO: OMG MOCK THIS IN OTHER TESTS
    datasets_list = []

    if kwargs['datasets'] != '':
        try:
            for raw_dataset_string in kwargs['datasets'].split(','):
                try:
                    dataset_path, dataset_commit = raw_dataset_string.split(':')
                except ValueError as exception:
                    dataset_path = raw_dataset_string
                    #TODO: Allow NULL on the API
                    dataset_commit = "dummy"

                dataset = path_to_dataset(dataset_path, context=context)
                datasets_list.append(OrderedDict({'dataset': dataset['id'], 'git_commit_hash': dataset_commit}))
        except ValueError as exception:
            pass

    python_version = PYTHON_VERSION_DEFAULTS.get(kwargs['python_version'], kwargs['python_version'])

    framework_slug = kwargs['framework']

    package_manager = kwargs['package_manager']
    package_manager = 'conda' if package_manager == 'anaconda' else package_manager

    requirements = kwargs['requirements']
    default_requirements = "requirements.{}".format("txt" if package_manager == 'pip' else "yml")
    requirements = default_requirements if not requirements else requirements
    git_commit_hash = kwargs['commit']

    context = {
        'parameters':
        {
            "module": kwargs['module'],
            "package_path": kwargs['package_path'],
            "package_manager": package_manager,
            "requirements": requirements,
            "time_limit": kwargs['time_limit'],
            "repository": project_id,
            "module_arguments": module_arguments,
            "framework":
            {
                "slug": framework_slug
            },
            "python_version": python_version,
            "datasets_set": datasets_list,
            "git_commit_hash": git_commit_hash,
        },
        'meta':
        {
            "name": kwargs['name'],
            "description": kwargs['description'],
        }
    }

    return context



from click.testing import CliRunner

from clusterone import ClusteroneClient
from clusterone.persistance.session import Session
from clusterone.client_exceptions import RemoteAquisitionFailure
from clusterone.clusterone_cli import cli
from clusterone.commands.create.dataset import cmd

DATASET_RESPONSE_LITERAL = {
    "display_name": "someDatasetName",
    "name": "somedatasetname",
    "source": "gitlab",
    "description": "This is a sample project description",
}

def test_passing(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'save', autospec=True, return_value=None)
    ClusteroneClient.get_dataset = mocker.Mock(
        return_value={'git_auth_link': 'https://user:58459eb14@git.clusterone.com/user/somedatasetname.git'})
    ClusteroneClient.create_dataset = mocker.Mock(return_value=DATASET_RESPONSE_LITERAL)
    cmd.time = mocker.Mock()

    CliRunner().invoke(cli, ['create', 'dataset', 's3', 'someDatasetName', '--description', 'This is a sample project description'])

    ClusteroneClient.create_dataset.assert_called_with('someDatasetName', 's3', 'This is a sample project description')


def test_default(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'save', autospec=True, return_value=None)
    ClusteroneClient.get_dataset = mocker.Mock(
        return_value={'git_auth_link': 'https://user:58459eb14@git.clusterone.com/user/somedatasetname.git'})
    ClusteroneClient.create_dataset = mocker.Mock(return_value=DATASET_RESPONSE_LITERAL)
    cmd.time = mocker.Mock()

    CliRunner().invoke(cli, ['create', 'dataset', 'someDatasetName'])

    ClusteroneClient.create_dataset.assert_called_with('someDatasetName', 'gitlab', '')


def test_message(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'save', autospec=True, return_value=None)
    ClusteroneClient.get_dataset = mocker.Mock(
        return_value={'git_auth_link': 'https://user:58459eb14@git.clusterone.com/user/somedatasetname.git'})
    ClusteroneClient.create_dataset = mocker.Mock(return_value="someDatasetName")
    cmd.time = mocker.Mock()

    result = CliRunner().invoke(cli, ['create', 'dataset', 'whatever-really'])

    assert 'Dataset creating, this might take up to a minute' in result.output


def test_outputs_remote_url(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'save', autospec=True, return_value=None)
    mocker.patch.object(Session, '__init__', autospec=True, return_value=None)
    Session.config = mocker.Mock()
    Session.retry_count, Session.retry_interval = 1, 1
    Session.config.read = mocker.Mock(return_value='{"username": "a", "token": "b"}')
    ClusteroneClient.get_dataset = mocker.Mock(
        return_value={'git_auth_link': 'https://user:58459eb14@git.clusterone.com/user/somedatasetname.git'})
    ClusteroneClient.create_dataset = mocker.Mock(return_value=DATASET_RESPONSE_LITERAL)
    cmd.time = mocker.Mock()

    result = CliRunner().invoke(cli, ['create', 'dataset', 'whatever-really'])

    assert 'https://user:58459eb14@git.clusterone.com/user/somedatasetname.git' in result.output


def test_remote_aquisition_failure(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'save', autospec=True, return_value=None)
    mocker.patch.object(Session, '__init__', autospec=True, return_value=None)
    Session.config = mocker.Mock()
    Session.retry_count, Session.retry_interval = 1, 1
    Session.config.read = mocker.Mock(return_value='{"username": "a", "token": "b"}')
    ClusteroneClient.get_dataset = mocker.Mock(return_value={'git_auth_link': ''})
    ClusteroneClient.create_dataset = mocker.Mock(return_value=DATASET_RESPONSE_LITERAL)
    cmd.time = mocker.Mock()

    result = CliRunner().invoke(cli, ['create', 'dataset', 'whatever-really'])

    assert isinstance(result.exception, RemoteAquisitionFailure)



import json
import logging
import random
import string
from contextlib import suppress
from os import environ
from subprocess import run, PIPE
from time import sleep
from click import progressbar

from coreapi import Client

from .types import UserCredentials

JUST_SHELL_COMMAND = "python3 clusterone/clusterone_cli.py"


def _just(command):
    just_command = "{} {}".format(JUST_SHELL_COMMAND, command)
    return just_command


def call(command, asserted=True):
    command_list = command.split()
    result = run(command_list, stdout=PIPE)

    if asserted:
        exit_code = result.returncode
        assert exit_code == 0, "Command: {} failed".format(command)

    return result.stdout


def just_exec(command):
    logging.debug("calling: {}".format(command))
    return call(_just(command)).decode("utf-8")


def generate_random_credentials():
    salt = ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(16))

    username = "cluster-one-test-{}".format(salt)
    password = ''.join(random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits) for _ in range(32))
    email = '{}@example.com'.format(salt)

    logging.info("generated username: {}".format(username))
    logging.info("generated pass: {}".format(password))

    return UserCredentials(username=username, password=password, email=email)


def authorize_client_via_session_file(client=None):
    """
    This is a hax to perform actions on a system that are not possible via CLI
    """

    with open("{}/.config/clusterone/session.json".format(environ['HOME'])) as session:
        token = json.loads("".join(session.readlines()))['token']

    client.token = token
    client.transport = client.get_transport()
    client.api_client = Client(
        transports=[client.transport], decoders=client.decoders)
    client.api_schema = client.download_schema()

    assert token is not None

    logging.info("token acquired: {}...".format(token[0:10]))


def poll(function, args=(), kwargs=None, timeout=None, interval=None):
    """
    Sometimes we don't know when the results shall be available, we only have a time frame
    """

    kwargs = kwargs or {}

    logging.info("polling for {} with interval of {}s, timeout is: {}s".format(function.__name__, interval, timeout))

    max_tries = int(timeout / interval)

    with progressbar(range(max_tries + 1), label="{}".format(function.__name__)) as try_count:
        for tries in try_count:
            kwargs = kwargs or {}

            with suppress(AssertionError):
                function(*args, **kwargs)
                logging.info("completed after {} retries".format(tries))
                break

            sleep(interval)
        else:
            raise AssertionError("Polling failed for {}({}, {})".format(function.__name__, ",".join(map(str, args)), ",".join(["{}={}".format(key, value) for key, value in zip(kwargs.keys(), kwargs.values())])))



# -*- coding: utf-8 -*-

import click

from clusterone import config
from clusterone.persistance.config import CONFIGURABLE_PROPERTIES
from clusterone.utilities import Choice
from clusterone.utilities import append_to_docstring


@click.command()
@click.pass_obj
@click.argument(
    'key',
    type=Choice(CONFIGURABLE_PROPERTIES),
    )
@click.argument(
    'value',
    )
@append_to_docstring("Available keys: {}".format(", ".join(CONFIGURABLE_PROPERTIES)))
def command(context, key, value):
    """
    Configure the CLI. Config is persistent.

    """

    setattr(config, key, value)



from collections import OrderedDict

from click.testing import CliRunner

from clusterone import ClusteroneClient
from clusterone.clusterone_cli import cli
from clusterone.commands.rm.job import cmd
from clusterone.client_exceptions import RunningJobRemoveAttempt


def test_deletion(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    ClusteroneClient.delete_job = mocker.Mock()
    cmd.serialize_job = mocker.Mock(return_value=OrderedDict([
        ('job_id', 'some-dummy-id'),
        ('status', 'created'),
    ]))

    mocker.spy(cmd, 'confirm')

    CliRunner().invoke(cli, ['rm', 'job', 'project/job_name'], 'y')

    cmd.serialize_job.assert_called_with('project/job_name', context=mocker.ANY)
    cmd.confirm.assert_called_with(mocker.ANY, abort=True)
    ClusteroneClient.delete_job.assert_called_with('some-dummy-id')

def test_running_job(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    ClusteroneClient.delete_job = mocker.Mock()
    cmd.serialize_job = mocker.Mock(return_value=OrderedDict([
        ('job_id', 'some-dummy-id'),
        ('status', 'running'),
    ]))

    result = CliRunner().invoke(cli, ['rm', 'job', 'project/job_name'], 'y')

    assert isinstance(result.exception, RunningJobRemoveAttempt)



import click
from click import confirm

from clusterone import authenticate
from clusterone.client_exceptions import BusyDatasetRemoveAttempt
from clusterone.utilities import path_to_dataset


@click.command()
@click.pass_obj
@authenticate()
@click.argument(
    'dataset-path',
    )
def command(context, dataset_path):
    """
    Removes a dataset, confirmation required
    """

    client = context.client

    dataset = path_to_dataset(dataset_path, context=context)

    # test
    jobs_using_this_dataset = client.get_jobs({
        "datasets": dataset['id'],
    })
    dataset_used_by_running_job = any(
        map(lambda job: job['status'] in ['started', 'running'],
            jobs_using_this_dataset)
    )
    if dataset_used_by_running_job:
        raise BusyDatasetRemoveAttempt()

    confirm("Dummy?", abort=True)

    client.delete_dataset(dataset['name'], dataset['owner']['username'])



from subprocess import call
import pytest

@pytest.mark.skip(reason="Cannot be run without cleanup, and the delete command isin't functional yet")
def test_comands_can_be_composed_together():

    #TODO: Make it run in the approperiate directory regardless of invocation path
    #TODO: Update the readme with caution info about the sponge package

    bashCommand = "python3 clusterone/clusterone_cli.py create project a26 | sponge | tail -n 1 | python3 clusterone/clusterone_cli.py ln project"
    exit_code = call(bashCommand, shell=True)

    print("Be sure to have install `sponge` package")
    assert exit_code is 0




import os
import re

import click
from click import Choice, Path
from click import IntRange
from click.exceptions import BadParameter

from clusterone import run_tf

try:
    from math import inf
except ImportError as exception:
    inf = float('inf')

# TODO: Move this to utilities
POSITIVE_INTEGER = IntRange(1, inf)


def validate_module(ctx, param, value):
    value = value.replace(".py", "")

    return value


def validate_mode(ctx, param, value):
    return "single-node" if value == "single" else value


def validate_env(ctx, param, value):
    return False if value == "new" else True


def validate_package_path(ctx, param, value):
    # default case
    if value == ".":
        return ""
    else:
        escaped_path = value.rstrip("/")

        path_components = escaped_path.split("/")
        module_path = ".".join(path_components)

        return module_path


@click.command()
@click.pass_context
@click.argument(
    "mode",
    type=Choice(["single", "distributed"]),
    callback=validate_mode,
)
@click.option(
    "--module",
    type=Path(),
    default="./main.py",
    callback=validate_module,
    help="Name of module to run. Default: \"main\"",
)
@click.option(
    # TODO: This will fail for more than 1 directory depth, fix in the future
    "--package-path",
    type=Path(exists=True),
    default=".",
    callback=validate_package_path,
    help="Path to module within the project. Default: project root",
)
@click.option(
    "--worker-replicas",
    type=POSITIVE_INTEGER,
    default=2,
    help="Number of worker instances. Default: 2",
)
@click.option(
    "--ps-replicas",
    type=POSITIVE_INTEGER,
    default=1,
    help="Number of parameter server instances. Default: 1",
)
@click.option(
    "--requirements",
    type=Path(exists=True),
    default="./requirements.txt",
    help="Requirements file to use. Default: requirements.txt",
)
@click.option(
    "--env",
    type=Choice(["current", "new"]),
    default="new",
    callback=validate_env,
    help="Environment to run the job in - 'new' rebuilds the environment from scratch"
         ", 'current' reuses the current environment. Default: new",
)
def command(context, mode, module, env, requirements, worker_replicas, ps_replicas, package_path):
    """
    Run job locally on a simulated server environment.

    MODE < single | distributed >
    """

    run_tf(
        cwd=os.getcwd(),
        package_path=package_path,
        module=module,
        mode=mode,
        worker_replicas=worker_replicas,
        ps_replicas=ps_replicas,
        requirements=requirements,
        current_env=env,
    )



import requests
from distutils.version import LooseVersion

import clusterone

PYPI_VERSIONS_URL = 'https://pypi.python.org/pypi/clusterone/json'

# Pypi converts "-alpha" -> "a", and "-beta" to "b"
PRERELEASE_SUFFIXES = ['dev', 'a', 'b', 'rc']


def is_latest_version():
    """
    Performs version check against Pypi
    """
    # TODO: Move this to function signature after removing 2.7 compliance
    # type: () -> bool

    remote_stable_versions = [LooseVersion(version) for version in get_pypi_versions() if is_stable(version)]
    remote_latest_stable = max(remote_stable_versions)

    current = LooseVersion(clusterone.__version__)

    return current >= remote_latest_stable


def is_stable(version): return not any(suffix in version for suffix in PRERELEASE_SUFFIXES)
# TODO: Move this to function signature after removing 2.7 compliance
# type: str -> bool


def get_pypi_versions():
    # TODO: Move this to function signature after removing 2.7 compliance
    # type: () -> [str]

    response_data = requests.get(PYPI_VERSIONS_URL).json()
    return response_data["releases"].keys()



import clusterone
from clusterone.mocks import PYPI_RESPONSE_LITERAL_V0DOT10DOT3

from . import version

PYPI_RESPONSE_LITERAL = {'releases': {'1234567890.0.0': [{}]}}
version.PYPI_VERSIONS_URL = "//some/url/to/pypi"
is_latest_version = version.is_latest_version


def test_up_to_date(mocker):
    clusterone.__version__ = "1234567891.30.51"
    json_mock = mocker.Mock()
    json_mock.json = mocker.Mock(return_value=PYPI_RESPONSE_LITERAL)
    version.requests.get = mocker.Mock(return_value=json_mock)

    assert is_latest_version()


def test_outdated(mocker):
    clusterone.__version__ = "20.0.0"

    is_latest_version = version.is_latest_version

    assert not is_latest_version()


def test_prereleases(mocker):
    """
    Even if there is later prerelease version only the latest stable should be taken into account
    """

    PYPI_RESPONSE_LITERAL = {'releases': {'1.0.2a10': [{}], '1.0.0': [{}]}}
    json_mock = mocker.Mock()
    json_mock.json = mocker.Mock(return_value=PYPI_RESPONSE_LITERAL)
    version.requests.get = mocker.Mock(return_value=json_mock)
    clusterone.__version__ = "1.0.1"

    assert is_latest_version()


def test_propper_sorting(mocker):
    """
    If lexical sorting is used than v9.0.0 would be before v10.0.0
    which is wrong

    Originated from CLUS-365.
    """

    json_mock = mocker.Mock()
    json_mock.json = mocker.Mock(return_value=PYPI_RESPONSE_LITERAL_V0DOT10DOT3)
    version.requests.get = mocker.Mock(return_value=json_mock)
    clusterone.__version__ = "0.10.0"

    assert not is_latest_version()



from collections import OrderedDict

from click.testing import CliRunner

from clusterone import ClusteroneClient
from clusterone.clusterone_cli import cli
from clusterone.commands.get.project import cmd


def test_passing_project_path(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    cmd.path_to_project = mocker.Mock()

    CliRunner().invoke(cli, ['get', 'project', 'some_user/some_project_name'])

    cmd.path_to_project.assert_called_with('some_user/some_project_name', context=mocker.ANY)

def test_table_call(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    cmd.path_to_project = mocker.Mock(return_value=OrderedDict([
        ('name', 'my-project-name'),
        ('owner', OrderedDict([
            ('username', 'someuser'),
            ])),
        ('created_at', '2020-06-12T11:35:56.364015Z'),
        ('modified_at', '2032-01-18T22:42:19.366416Z'),
        ('size', 123),
    ]))
    cmd.make_table = mocker.Mock(return_value="Example table string")

    CliRunner().invoke(cli, ['get', 'project', 'some_user/some_project_name'])

    cmd.make_table.assert_called_with([[
        'my-project-name',
        '2020-06-12T11:35:56.364015Z',
        'someuser',
        123,
        '2032-01-18T22:42:19.366416Z'
        ]], header=['Name', 'Created', 'Owner', 'Size [Mb]', 'Modified'])

def test_printing_out_table(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    cmd.path_to_project = mocker.Mock(return_value=OrderedDict([
        ('name', 'my-project-name'),
        ('owner', OrderedDict([
            ('username', 'someuser'),
            ])),
        ('created_at', '2020-06-12T11:35:56.364015Z'),
        ('modified_at', '2032-01-18T22:42:19.366416Z'),
        ('size', 123),
    ]))
    cmd.make_table = mocker.Mock(return_value="Example table string")

    result = CliRunner().invoke(cli, ['get', 'project', 'some_user/some_project_name'])

    assert result.output == "Example table string\n"



"""
Clusterone Click types for use in commands

This is essentially a bridge between Clusterone Python client and Click.
"""
from click import ParamType

from clusterone.just_client import Notebook as JustNotebook
from clusterone import client


class Notebook(ParamType):
    name = "jupyter notebook"

    def convert(self, value, param, ctx):
        return JustNotebook.from_clusterone(client, value)

    def __repr__(self):
        return "Notebook"



"""
Convention about keys: please use lowercase only
Please make sure to mark configurable properties with is_configurable decorator
"""
# TODO: Move this to general Config

import json
import os

import click
import py
from click.exceptions import BadParameter
from coreapi.exceptions import NetworkError

from clusterone.just_client import ClusteroneClient
from .defaults import CONFIG_DEFAULTS

CONFIGURABLE_PROPERTIES = []
# TODO: Export click Choice type in the future (well, actually Choice from clusterone.utilities)

def is_configurable(function):
    """
    Indicates that this property can be configured by the user
    """

    CONFIGURABLE_PROPERTIES.append(function.__name__)

    return function

class Config(object):
    def __init__(self, defaults=CONFIG_DEFAULTS):
        #TODO: When dropping python2 compliacne add * to function signature

        self.file = py.path.local(
            click.get_app_dir('clusterone')).join('justrc.json')

        self.defaults = defaults

    # if file does not exists create it
    # if file is corrupted and cannot be read / written inform user about this
    #TODO: TEST THIS!!!!!!!!!!

    def __setattr__(self, key, value):
        """
        Normalizes the keys to lowercase only
        """

        return object.__setattr__(self, key.lower(), value)

    # redo this to protected methods
    # do a generic __setattr__ and __getattr__
    def set(self, key, value):
        #TODO: Docstring

        #TODO: Smarter somethign
        #TODO: Check for existing solution
        #TODO: Context manager for json
        #TODO: Don't load and reqrite whole file, do it in a smarter way
        #TODO: Extract to utilities

        dir_path = os.path.dirname(str(self.file))
        if not os.path.exists(dir_path):
            os.mkdir(dir_path)

        with self.file.open('w+', encoding='utf-8') as justrc:

            try:
                json_content = json.load(justrc)
            except ValueError:
                json_content = {}

            json_content[key] = value

            try:
                # pretty printing is important for the config to be human editable
                justrc.write(json.dumps(json_content, indent=4, sort_keys=True))
            except TypeError:
                # Python 2 compliance - .write() required unicode
                justrc.write(json.dumps(json_content).decode('utf-8'))

    def get(self, key):
        # TODO: OMG, refactor like shit!!!!!!!
        try:
            with self.file.open('r', encoding='utf-8') as justrc:

                try:
                    json_content = json.load(justrc)
                except ValueError:
                    json_content = {}

                return json_content[key]
        except (KeyError, py.error.ENOENT):
            return self.defaults[key]


    #TODO: Can this be done dynamically?
    # yup - __getattr__ or __getattribute__

    #TODO: Move validation from cmd here
    @property
    @is_configurable
    def endpoint(self):
        """
        the URL to Clusterone service
        """

        # TODO: Redo this to [] key aquisition
        return self.get('endpoint')

    @endpoint.setter
    def endpoint(self, value):
        #type (object, str) -> str
        # TODO: Move this to function signature after dropping Python 2.7 compliance

        if not value.endswith('/'):
            value = "{}/".format(value)

        if "api" not in value:
            value = "{}api/".format(value)

        try:
            ClusteroneClient(api_url=value)
        except NetworkError:
            raise BadParameter("please make sure it's a valid URL pointing to Clusterone service.", param_hint="endpoint")

        self.set('endpoint', value)



"""
Convert time_limit_string string [ex. 22h5m] to total minute count [ex. 1325]
"""

import re as regexp

MINUTES_IN_HOUR = 60
TIME_LIMIT_REGEXP = "\\b(\\d+)h\\b|\\b(\\d+)m\\b|\\b(\\d+)h(\\d+)m\\b"

def main(time_limit_string):

    match = regexp.match(TIME_LIMIT_REGEXP, time_limit_string, flags=regexp.IGNORECASE)

    if not match:
        raise ValueError("Invalid time_limit_string, regexp did not match")

    all_groups = match.groups()
    groups = all_groups[:2] if all_groups[1] or all_groups[0] else all_groups[2:4]
    groups = tuple(map(lambda group: int(group or 0), groups))
    hours, minutes = groups

    return hours * MINUTES_IN_HOUR + minutes



# -*- coding: utf-8 -*-

import gc

import click
from click import echo

from clusterone.utilities import is_latest_version
from clusterone.client_exceptions import NotSupported

def perform_version_check():
    """
    Prevents any operation on the CLI if it's not up to date
    """

    if not is_latest_version():
        raise NotSupported()

@click.command()
@click.option('--username', '-u', prompt=True)
@click.option('--password', '-p', prompt=True, hide_input=True)
#TODO: Since Click 7.0 there is an show=False keyword avaible, it should be used here
# This is a hack to run version check before the credential prompt
@click.option('--version-validator', is_eager=True, callback=lambda context, param, value: perform_version_check(), help="Please ignore this parameter")
@click.pass_obj
def command(context, username, password, version_validator):
    """
    Log into Clusterone
    """

    client, session = context.client, context.session

    client.api_login(username, password)

    # purgin password literal from memory
    del password
    gc.collect()

    session['token'] = client.token
    session['git_token'] = client.git_token
    session['username'] = client.username

    session.save()
    echo("Login successful")



from click import prompt

def silent_prompt(**kwargs):
    return prompt('', prompt_suffix='', **kwargs)

from random import choice, randint

#TODO: Allow this stuff as session? :D
MIN_NUM = 100
MAX_NUM = 999
adjs = [
    "autumn", "hidden", "bitter", "misty", "silent", "empty", "dry", "dark",
    "summer", "icy", "delicate", "quiet", "white", "cool", "spring", "winter",
    "patient", "twilight", "dawn", "crimson", "wispy", "weathered", "blue",
    "billowing", "broken", "cold", "damp", "falling", "frosty", "green", "long",
    "late", "lingering", "bold", "little", "morning", "muddy", "old", "red",
    "rough", "still", "small", "sparkling", "throbbing", "shy", "wandering",
    "withered", "wild", "black", "young", "holy", "solitary", "fragrant",
    "aged", "snowy", "proud", "floral", "restless", "divine", "polished",
    "ancient", "purple", "lively", "nameless"
]
nouns = [
    "waterfall", "river", "breeze", "moon", "rain", "wind", "sea", "morning",
    "snow", "lake", "sunset", "pine", "shadow", "leaf", "dawn", "glitter",
    "forest", "hill", "cloud", "meadow", "sun", "glade", "bird", "brook",
    "butterfly", "bush", "dew", "dust", "field", "fire", "flower", "firefly",
    "feather", "grass", "haze", "mountain", "night", "pond", "darkness",
    "snowflake", "silence", "sound", "sky", "shape", "surf", "thunder",
    "violet", "water", "wildflower", "wave", "water", "resonance", "sun",
    "wood", "dream", "cherry", "tree", "fog", "frost", "voice", "paper", "frog",
    "smoke", "star"
]

def random_job_name():
    return "-".join(
        [choice(adjs),
        choice(nouns),
        str(randint(MIN_NUM, MAX_NUM))])




"""
This module is suplementing tests for Choice utility.
Only added functionality is tested because of inheritance.
"""

from .choice import Choice, lazify, lazy_property


def test_generator_evaluation(mocker):
    """
    Sometimes the choice is based on dynamic data that takes a lot of time to aquire, it should be aquired iff necessary.
    """

    long_running_function = mocker.Mock(return_value=['1', '2', '3'])

    def sample_generator():
        for value in long_running_function():
            yield value

    choice = Choice(sample_generator())
    assert not long_running_function.called

    assert choice.choices == ['1','2','3']
    assert long_running_function.called

def test_generator_caching():
    """
    Check if the generator iterator is not exhaused internally but propperly cached.
    """

    def sample_generator():
        for value in ['dummy', 'bunny', 'dummy_bunny']:
            yield value

    choice = Choice(sample_generator())

    # if not cached this would exhaust the iterator
    choice.choices

    assert choice.choices == ['dummy', 'bunny', 'dummy_bunny']


def test_lazify(mocker):

    long_running_function = mocker.Mock(return_value=[1, 2, 3])

    @lazify
    def test_function():
        return long_running_function()

    generator_instance = test_function()
    assert not long_running_function.called

    assert list(generator_instance) == [1, 2, 3]
    assert long_running_function.called


def test_lazy_property():

    RETVAL = [1, 2, 3]

    class Sample():
        @property
        @lazify
        def dummy(self):
            return RETVAL


    class Test():
        @lazy_property
        def dummy(self):
            return RETVAL

    sample, test = Sample(), Test()

    assert list(sample.dummy) == list(test.dummy)



from click.testing import CliRunner

from clusterone import ClusteroneClient
from clusterone.persistance.session import Session
from clusterone.clusterone_cli import cli

from clusterone.commands.matrix import cmd


def test_opening(mocker):
    """
    Call to click utility for browser control
    """

    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    mocker.patch.object(Session, 'save', autospec=True, return_value=None)
    ClusteroneClient.matrix_url = "https://dummy/matrix"

    cmd.launch = mocker.Mock()

    CliRunner().invoke(cli, ['matrix'],)

    cmd.launch.assert_called_with("https://dummy/matrix")



import click
from click import confirm

from clusterone import authenticate
from clusterone.client_exceptions import BusyProjectRemoveAttempt
from clusterone.utilities import path_to_project

@click.command()
@click.pass_obj
@click.argument(
    'project-path-or-id',
    )
@authenticate()
def command(context, project_path_or_id):
    """
    Removes a project, confirmation required
    """

    client = context.client

    project = path_to_project(project_path_or_id, context=context)

    project_jobs = client.get_jobs({
        "repository": project['id'],
    })
    project_has_running_jobs = any(
        map(lambda job: job['status'] in ['started', 'running'], project_jobs)
    )

    if project_has_running_jobs:
        raise BusyProjectRemoveAttempt()

    confirm("Removing a project will remove all associated jobs and outputs. Are you sure?", abort=True)

    client.delete_project(project['name'], project['username'])



import click

from clusterone import authenticate
from clusterone.utilities import path_to_project, make_table


HEADER = ['Name', 'Created', 'Owner', 'Size [Mb]', 'Modified']

def extract_data_from_project(project):
    return [
        project['name'],
        project['created_at'],
        project['owner']['username'],
        project['size'],
        project['modified_at'],
    ]

@click.command()
@click.pass_obj
@authenticate()
@click.argument(
    'project_path',
    )
def command(context, project_path):
    """
    Get information about a project
    """

    project = path_to_project(project_path, context=context)
    project_data = extract_data_from_project(project)
    click.echo(make_table([project_data], header=HEADER))

    return project



from click.testing import CliRunner

from clusterone.clusterone_cli import cli
from clusterone import ClusteroneClient
from clusterone.commands.create.notebook import cmd
from clusterone.mocks import GET_FRAMEWORKS_RESPONSE, GET_INSTANCE_TYPES_RESPONSE


# base_options call is not explicitly tested as other tests depend on that call

def test_passing_instance_type(mocker):
    # This mocks will propagate across the tests
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    ClusteroneClient.create_job = mocker.Mock()

    ClusteroneClient.get_frameworks = mocker.Mock(return_value=GET_FRAMEWORKS_RESPONSE)
    ClusteroneClient.get_instance_types = mocker.Mock(return_value=GET_INSTANCE_TYPES_RESPONSE)

    ClusteroneClient.create_job = mocker.Mock()
    cmd.base = mocker.Mock(return_value={'meta': {'name': 'late-moon-758', 'description': ''}, 'parameters': {'package_path': '', 'datasets_set': [], 'requirements': 'requirements.txt', 'time_limit': 2880, 'module': 'main', 'framework': 'tensorflow-1.3.0', 'git_commit_hash': '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1', 'code': 'aaf4de71-f506-48c0-855c-02c7c485c5a4', 'package_manager': 'pip', 'Clusterone_api_token': 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJlbWFpbCI6Im9sZ2llcmRAa2FzcHJvd2ljei5wcm8iLCJ1c2VyX2lkIjo3OTUsImV4cCI6MTUxNzMxNjUzMCwidXNlcm5hbWUiOiJhbGxncmVlZCJ9.IJhEZWwMYf2sjHhoxUsjCj0Xll5CVX-RO3eUqvH7myU', 'python_version': 2.7}})
    cmd.Notebook.__init__ = mocker.Mock()

    CliRunner().invoke(cli, [
        'create',
        'notebook',
        '--project', 'someproject',
        '--instance-type', 'p2.xlarge',
    ])

    args, kwargs = cmd.Notebook.__init__.call_args
    assert args[1]['workers']['slug'] == 'p2.xlarge'


def test_default_instance_type(mocker):
    cmd.client = mocker.Mock()
    cmd.base = mocker.Mock(return_value={'meta': {'name': 'late-moon-758', 'description': ''}, 'parameters': {'package_path': '', 'requirements': 'requirements.txt', 'time_limit': 2880, 'module': 'main', 'tf_version': '', 'framework': 'tensorflow', 'git_commit_hash': '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1', 'code': 'aaf4de71-f506-48c0-855c-02c7c485c5a4', 'package_manager': 'pip', 'Clusterone_api_token': 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJlbWFpbCI6Im9sZ2llcmRAa2FzcHJvd2ljei5wcm8iLCJ1c2VyX2lkIjo3OTUsImV4cCI6MTUxNzMxNjUzMCwidXNlcm5hbWUiOiJhbGxncmVlZCJ9.IJhEZWwMYf2sjHhoxUsjCj0Xll5CVX-RO3eUqvH7myU', 'python_version': 2.7}})
    cmd.Notebook.__init__ = mocker.Mock()

    CliRunner().invoke(cli, [
        'create',
        'notebook',
        '--project', 'someproject',
    ])

    args, kwargs = cmd.Notebook.__init__.call_args
    assert args[1]['workers']['slug'] == 't2.small'


def test_call_to_base(mocker):
    cmd.client = mocker.Mock()
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    cmd.base = mocker.Mock()
    cmd.Notebook.__init__ = mocker.Mock()

    CliRunner().invoke(cli, [
        'create',
        'notebook',
        '--project', 'someproject',
    ])

    assert cmd.base.call_count == 1


def test_is_single(mocker):
    cmd.client = mocker.Mock()
    cmd.base = mocker.Mock(return_value={'meta': {'name': 'late-moon-758', 'description': ''}, 'parameters': {'package_path': '', 'requirements': 'requirements.txt', 'time_limit': 2880, 'module': 'main', 'tf_version': '', 'framework': 'tensorflow', 'git_commit_hash': '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1', 'code': 'aaf4de71-f506-48c0-855c-02c7c485c5a4', 'package_manager': 'pip', 'Clusterone_api_token': 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJlbWFpbCI6Im9sZ2llcmRAa2FzcHJvd2ljei5wcm8iLCJ1c2VyX2lkIjo3OTUsImV4cCI6MTUxNzMxNjUzMCwidXNlcm5hbWUiOiJhbGxncmVlZCJ9.IJhEZWwMYf2sjHhoxUsjCj0Xll5CVX-RO3eUqvH7myU', 'python_version': 2.7}})
    cmd.Notebook.__init__ = mocker.Mock()

    CliRunner().invoke(cli, [
        'create',
        'notebook',
        '--project', 'someproject',
    ])

    args, kwargs = cmd.Notebook.__init__.call_args
    assert args[1]['mode'] == 'single'
    assert args[1]['framework']['slug'] == 'jupyter'


def test_is_client_called(mocker):
    cmd.client = mocker.Mock()
    cmd.base = mocker.Mock(return_value={'meta': {'name': 'late-moon-758', 'description': ''}, 'parameters': {'package_path': '', 'requirements': 'requirements.txt', 'time_limit': 2880, 'module': 'main', 'tf_version': '', 'framework': 'tensorflow', 'git_commit_hash': '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1', 'code': 'aaf4de71-f506-48c0-855c-02c7c485c5a4', 'package_manager': 'pip', 'Clusterone_api_token': 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJlbWFpbCI6Im9sZ2llcmRAa2FzcHJvd2ljei5wcm8iLCJ1c2VyX2lkIjo3OTUsImV4cCI6MTUxNzMxNjUzMCwidXNlcm5hbWUiOiJhbGxncmVlZCJ9.IJhEZWwMYf2sjHhoxUsjCj0Xll5CVX-RO3eUqvH7myU', 'python_version': 2.7}})
    cmd.Notebook.__init__ = mocker.Mock()

    CliRunner().invoke(cli, [
        'create',
        'notebook',
        '--project', 'someproject',
    ])

    assert cmd.Notebook.__init__.called


def test_outputting_notebook_url_and_token(mocker):
    cmd.client = mocker.Mock()
    cmd.base = mocker.Mock(return_value={'meta': {'name': 'late-moon-758', 'description': ''}, 'parameters': {'package_path': '', 'requirements': 'requirements.txt', 'time_limit': 2880, 'module': 'main', 'tf_version': '', 'framework': 'tensorflow', 'git_commit_hash': '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1', 'code': 'aaf4de71-f506-48c0-855c-02c7c485c5a4', 'package_manager': 'pip', 'Clusterone_api_token': 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJlbWFpbCI6Im9sZ2llcmRAa2FzcHJvd2ljei5wcm8iLCJ1c2VyX2lkIjo3OTUsImV4cCI6MTUxNzMxNjUzMCwidXNlcm5hbWUiOiJhbGxncmVlZCJ9.IJhEZWwMYf2sjHhoxUsjCj0Xll5CVX-RO3eUqvH7myU', 'python_version': 2.7}})
    mock_notebook = mocker.Mock()
    mock_notebook.url = "my_dummy_url"
    mock_notebook.token = "token"
    cmd.Notebook.__new__ = mocker.Mock(return_value=mock_notebook)

    result = CliRunner().invoke(cli, [
        'create',
        'notebook',
        '--project', 'someproject',
    ])

    assert result.output == "my_dummy_url\naccess token: token\n"



import os
import argparse
import subprocess
import sys
import time
import atexit
import signal
from multiprocessing import Pool, Process
import shutil
import stat

# List that stores all the children processes of this module
jobs = []
processes = []


def cleanup():
    """ Kills all the children processes.
    """
    for job in jobs:
        job.terminate()
        job.join()


# Register cleanup function to the exit of this module
atexit.register(cleanup)


def which(pgm):
    path = os.getenv('PATH')
    for p in path.split(os.path.pathsep):
        p = os.path.join(p, pgm)
        if os.path.exists(p) and os.access(p, os.X_OK):
            return p


def build_env(name, requirements=None, current_env=False, tf_version=''):
    if current_env:
        try:
            pybin = os.path.join(os.environ['VIRTUAL_ENV'], 'bin/python')
        except:
            pybin = sys.executable
    else:
        virtualenv_path = os.path.expanduser("~/.clusterone/env/%s/" % name)
        env_script_path = os.path.expanduser("~/.clusterone/env/%s/bootstrap.sh" % name)
        try:
            os.makedirs(virtualenv_path)
        except:
            shutil.rmtree(virtualenv_path)
            os.makedirs(virtualenv_path)
        activate_cmd = os.path.join(virtualenv_path, 'bin/activate')
        pybin = os.path.join(virtualenv_path, 'bin/python')
        virtualenv_bin = which('virtualenv')

        if tf_version is not '':
            tf_version = '==%s' % tf_version

        with open(env_script_path, 'w') as f:
            f.write("#!/bin/bash\n")
            f.write("%s %s\n" % (virtualenv_bin, virtualenv_path))
            f.write("source %s\n" % (activate_cmd))
            f.write("pip install tensorflow%s\n" % (tf_version))
            f.write("pip install clusterone\n")
            if requirements is not None:
                requirements_cmd = 'pip install -r %s' % requirements
                f.write("pip install -r %s\n" % (requirements))

        st = os.stat(env_script_path)
        os.chmod(env_script_path, st.st_mode | stat.S_IEXEC | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
        p = subprocess.Popen("%s" % (env_script_path), shell=True)
        (output, error) = p.communicate()
    return pybin


def run_cmd(pybin, cmd, cwd, env={}):
    """
        run command in shell and captures output.
        args:
            name: string,
            cmd: string, the command.
            cwd: string, working directory for the command.
            env: dictionary, enviroment variables
    """

    env_cmd = '%s %s' % (pybin, cmd)
    signal.signal(signal.SIGINT, signal_handler)
    p = subprocess.Popen(env_cmd, cwd=cwd, env=env, shell=True)
    processes.append(p)
    (output, error) = p.communicate()
    if error:
        print("%s - ERROR: " % error)


def signal_handler(signal, frame):
    for p in processes:
        p.terminate()
    sys.exit(0)


def run_tf(cwd,
           package_path,
           module,
           mode,
           worker_replicas,
           ps_replicas,
           requirements=None,
           current_env=False,
           tf_version=''):
    assert (not module == '')

    # If package path exists then runs package_path.module else runs module only
    if package_path:
        package_path = '%s.' % package_path
    else:
        package_path = ''

    pybin = build_env('test',
                      requirements=requirements,
                      current_env=current_env,
                      tf_version=tf_version)

    # Single mode
    if mode == 'single-node':
        cmd = '-m %s%s' % (package_path, module)
        run_cmd(pybin, cmd, cwd, {})

    # Distributed mode
    if mode == 'distributed':
        # Compute ps_hosts string, assuming ps_hosts serve on ports 11000+
        ps_hosts = "localhost:11000"
        for i in range(ps_replicas - 1):
            ps_hosts += ",localhost:%d" % (11000 + i + 1)

        # Compute worker_hosts string, assuming worker_hosts serve on ports 12000+
        worker_hosts = "localhost:12000"
        for i in range(worker_replicas - 1):
            worker_hosts += ",localhost:%d" % (12000 + i + 1)

        # Spin up workers
        for i in range(worker_replicas):
            cmd = '-m %s%s' % (package_path, module)
            env = {
                "JOB_NAME": "worker",
                "TASK_INDEX": str(i),
                "PS_HOSTS": ps_hosts,
                "WORKER_HOSTS": worker_hosts
            }
            p = Process(target=run_cmd, args=(pybin, cmd, cwd, env,))
            p.daemon = True
            p.start()
            jobs.append(p)

        # Spin up parameter servers
        for i in range(ps_replicas):
            cmd = '-m %s%s' % (package_path, module)
            env = {
                "JOB_NAME": "ps",
                "TASK_INDEX": str(i),
                "PS_HOSTS": ps_hosts,
                "WORKER_HOSTS": worker_hosts
            }
            p = Process(target=run_cmd, args=(pybin, cmd, cwd, env,))
            p.daemon = True
            p.start()
            jobs.append(p)

        for job in jobs:
            job.join()


if __name__ == "__main__":
    """
        python tf_runner.py -module mnist -mode distributed -worker_replicas 2 -ps_replicas 1
    """
    parser = argparse.ArgumentParser(description='Run TF experiment')
    parser.add_argument('--module', type=str, help="Python module")
    parser.add_argument(
        '--package-path', type=str, default=None, help="Python package path")
    parser.add_argument(
        '--mode', type=str, default='single', help="distributed or single-node")
    parser.add_argument(
        '--worker-replicas',
        type=int,
        default=0,
        help="Number of worker replicas")
    parser.add_argument(
        '--ps-replicas',
        type=int,
        default=0,
        help="Number of parameter server replicas")
    parser.add_argument(
        '--requirements',
        type=int,
        default=0,
        help="Number of parameter server replicas")
    args = parser.parse_args()
    assert (args.mode in ('single-node', 'distributed'))
    run_tf(
        cwd=os.getcwd(),
        package_path=args.package_path,
        module=args.module,
        mode=args.mode,
        worker_replicas=args.worker_replicas,
        ps_replicas=args.ps_replicas)



import click

from clusterone import authenticate
from clusterone.utilities import path_to_dataset, make_table


HEADER = ['Property', 'Value']

HEADER_BASE = [
    'Name',
    'Id',
    'Source',
    'URL'
]

def extract_data_from_dataset(dataset):
    extracted_data = [
        dataset['name'],
        dataset['id'],
        dataset['source'],
        dataset['http_url_to_repo'],
    ]

    key_value_pairs = zip(HEADER_BASE, extracted_data)

    return [pair for pair in key_value_pairs]

@click.command()
@click.pass_obj
@authenticate()
@click.argument(
    'dataset_path',
    )
def command(context, dataset_path):
    """
    Get information about a dataset
    """

    dataset = path_to_dataset(dataset_path, context=context)
    dataset_data = extract_data_from_dataset(dataset)
    click.echo(make_table(dataset_data, header=HEADER))

    return dataset



def append_to_docstring(text):
    """
    Potentially usefull with Click commands that utilze docstrings as help messages
    """

    def wrapper(function):
        function.__doc__ += text
        return function
    return wrapper



from coreapi.document import Error
from coreapi.exceptions import ErrorMessage
from pytest import raises

from clusterone.client_exceptions import _NonExistantNotebook, NonExistantNotebook, DuplicateNotebookName
from clusterone.just_client.types import TaskStatus
from clusterone.mocks import NOTEBOOK_JOB_CONFIGURATION, NOTEBOOK_API_RESPONSE
from .notebook import Notebook, NotebookPath


def test_existing_from_uuid(mocker):
    mock_client = mocker.Mock()
    mock_client.client_action.return_value = NOTEBOOK_API_RESPONSE

    target = Notebook._from_data(NOTEBOOK_API_RESPONSE)

    notebook = Notebook.from_clusterone(mock_client, "853b9f10-36ce-4de4-b2f8-108d69733b42")
    args, kwargs = mock_client.client_action.call_args

    assert notebook == target
    assert args[0] == ['notebooks', 'read']
    assert kwargs['params'] == {
        'job_id': "853b9f10-36ce-4de4-b2f8-108d69733b42",
    }


def test_existing_from_path(mocker):
    mock_client = mocker.Mock()
    mock_client.client_action.return_value = NOTEBOOK_API_RESPONSE

    target = Notebook._from_data(NOTEBOOK_API_RESPONSE)

    notebook = Notebook.from_clusterone(mock_client, "dummy/winter-leaf-119")
    args, kwargs = mock_client.client_action.call_args

    assert notebook == target
    assert args[0] == ['notebook_by_name', 'read']
    assert kwargs['params'] == {
        "display_name": "winter-leaf-119",
        "username": "dummy",
    }


def test_existing_from_path_without_username(mocker):
    mock_client = mocker.Mock()
    mock_client.client_action.return_value = NOTEBOOK_API_RESPONSE
    mock_client.username = "dummy"

    Notebook.from_clusterone(mock_client, "winter-leaf-119")
    args, kwargs = mock_client.client_action.call_args

    assert kwargs['params']['username'] == "dummy"


def test_failed_acquisition_by_name(mocker):
    mock_client = mocker.Mock()
    mock_client.client_action.side_effect = ErrorMessage(
        Error(title='404 Not Found', content={'detail': 'Notebook not found'}))

    with raises(NonExistantNotebook):
        Notebook.from_clusterone(mock_client, "dummy/winter-leaf-119")


def test_failed_acquisition_by_id(mocker):
    mock_client = mocker.Mock()
    mock_client.client_action.side_effect = ErrorMessage(
        Error(title='404 Not Found', content={'detail': 'Notebook not found'}))

    with raises(_NonExistantNotebook):
        Notebook.from_clusterone(mock_client, "853b9f10-36ce-4de4-b2f8-108d69733b42")


def test_new(mocker):
    mock_client = mocker.Mock()
    mock_client.client_action.return_value = NOTEBOOK_API_RESPONSE

    target = Notebook._from_data(NOTEBOOK_API_RESPONSE)

    assert Notebook(mock_client, NOTEBOOK_JOB_CONFIGURATION) == target

    args, kwargs = mock_client.client_action.call_args
    assert args[0] == ['notebooks', 'create']
    assert kwargs['params'] == {
        'repository': None,
        'display_name': 'snowy-surf-115',
        'description': '',
        'parameters': NOTEBOOK_JOB_CONFIGURATION,
        'datasets_set': [],
    }


def test_failed_creation(mocker):
    mock_client = mocker.Mock()
    mock_client.client_action.side_effect = ErrorMessage(Error(title='400 Bad Request', content={
        ' messages': [' Fields username and display_name must make a unique set.']}))
    with raises(DuplicateNotebookName):
        Notebook(mock_client, NOTEBOOK_JOB_CONFIGURATION)


def test_initialisation():
    test_notebook = Notebook.__new__(Notebook)
    test_notebook._initialize(NOTEBOOK_API_RESPONSE)

    assert test_notebook.id == "853b9f10-36ce-4de4-b2f8-108d69733b42"
    assert test_notebook.url == "http://853b9f10-36ce-4de4-b2f8-108d69733b42.jupyter.v2.clusterone.com"
    assert test_notebook.status == TaskStatus.created
    assert test_notebook.token == '85c88d21-fb18-42be-8a1a-f73da4585a02'


def test_from_data_constructor():
    template = Notebook.__new__(Notebook)
    template._initialize(NOTEBOOK_API_RESPONSE)

    target = Notebook._from_data(NOTEBOOK_API_RESPONSE)

    assert template == target


def test_starting(mocker):
    mock_client = mocker.Mock()
    test_notebook = Notebook._from_data(NOTEBOOK_API_RESPONSE)

    test_notebook.start(mock_client)

    assert test_notebook.status == TaskStatus.started
    args, kwargs = mock_client.client_action.call_args
    assert args[0] == ['notebooks', 'start']
    assert kwargs['params'] == {
        'job_id': '853b9f10-36ce-4de4-b2f8-108d69733b42',
    }


def test_stopping(mocker):
    mock_client = mocker.Mock()
    test_notebook = Notebook._from_data(NOTEBOOK_API_RESPONSE)

    test_notebook.stop(mock_client)

    assert test_notebook.status == TaskStatus.stopped
    args, kwargs = mock_client.client_action.call_args
    assert args[0] == ['notebooks', 'stop']
    assert kwargs['params'] == {
        'job_id': '853b9f10-36ce-4de4-b2f8-108d69733b42',
    }


def test_repr():
    test_notebook = Notebook._from_data(NOTEBOOK_API_RESPONSE)

    assert repr(test_notebook) == "Notebook(<client_instance>, '853b9f10-36ce-4de4-b2f8-108d69733b42')"


def test_parsing_notebook_path():
    assert Notebook._parse_notebook_path("username/notebook") == NotebookPath(username="username", notebook_name="notebook")
    assert Notebook._parse_notebook_path("notebook") == NotebookPath(username=None, notebook_name="notebook")

    with raises(ValueError):
        Notebook._parse_notebook_path("keton/keton/keton/keton")

    with raises(ValueError):
        Notebook._parse_notebook_path("/")



from click.exceptions import BadParameter
from clusterone.client_exceptions import JobNameConflict, NonExistantJob

def path_to_project(project_path, context=None):
    """
    Coverts project_path [format. username/project] to ClusterOne project_id
    """

    client, session = context.client, context.session

    project_tokens = project_path.split('/')
    if len(project_tokens) == 1:
        project_tokens = [session.get("username")] + project_tokens

    if not len(project_tokens) == 2:
        raise BadParameter(param_hint="project path", message="Please provide a valid project path. The format is \"username/project\".")

    username, project_name = project_tokens
    project = client.get_project(project_name, username)
    return project

def path_to_job_id(job_path, context=None):
    """
    Coverts job_path [format. username/project/job] to ClusterOne job id
    """

    client, session = context.client, context.session

    path_tokens = job_path.split('/')

    if len(path_tokens) == 2:
        path_tokens = [session.get("username")] + path_tokens
    elif not len(path_tokens) == 3:
        raise BadParameter(param_hint="job-path", message="Please provide a valid job-path. The format is \"username/project/job-name\".")

    owner, project_name, job_name = path_tokens

    job_start_candidates = client.get_jobs({
        "repository": client.get_project(project_name, owner)['id'],
        "display_name": job_name
        })

    if len(job_start_candidates) is 0:
        raise NonExistantJob()

    if len(job_start_candidates) is not 1:
        candidate_ids = list(map(lambda job: job['job_id'], job_start_candidates))
        raise JobNameConflict(candidate_ids)

    job_id = job_start_candidates[0]['job_id']

    return job_id

def path_to_dataset(dataset_path, context=None):
    """
    Coverts dataset_path [format. username/dataset] to ClusterOne dataset_id
    """

    client, session = context.client, context.session

    dataset_tokens = dataset_path.split('/')
    if len(dataset_tokens) == 1:
        dataset_tokens = [session.get("username")] + dataset_tokens

    if not len(dataset_tokens) == 2:
        raise BadParameter(param_hint="dataset path", message="Please provide a valid dataset path. The format is \"username/dataset\".")

    username, dataset_name = dataset_tokens
    dataset = client.get_dataset(dataset_name, username)
    return dataset



import logging
from os import environ
from datetime import datetime

from .runner import run
from .getting_started_scenario import GettingStarted

LOG_PATH = "./logs"
LOG_FILENAME = datetime.utcnow().strftime("%Y-%m-%d_%H-%M-UTC")


log_path = "{0}/{1}.log".format(LOG_PATH, LOG_FILENAME)
log_formatter = logging.Formatter("%(asctime)s: [%(levelname)s] %(message)s", datefmt='%m/%d/%Y %H:%M:%S')

file_handler = logging.FileHandler(log_path)
file_handler.setFormatter(log_formatter)

console_handler = logging.StreamHandler()
console_handler.setFormatter(log_formatter)
console_handler.setLevel(logging.INFO)

the_logger = logging.getLogger()
the_logger.setLevel(logging.DEBUG)
the_logger.addHandler(file_handler)
the_logger.addHandler(console_handler)

logging.info("Logger init, saving to path: {}".format(log_path))

run(GettingStarted)



from click.testing import CliRunner

from clusterone import ClusteroneClient
from clusterone.clusterone_cli import cli

from clusterone.commands.start.job import cmd

def test_parameter_requirement(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)

    result = CliRunner().invoke(cli, ['start', 'job'])

    # This is a fun way of saying internal Click Exception
    assert str(result.exception) == '2'

def test_both_parameters_sepcified(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)

    result = CliRunner().invoke(cli, ['start', 'job', '--job-id', '12354', '--job-path', 'user/project/job'])

    # This is a fun way of saying internal Click Exception
    assert str(result.exception) == '2'

def test_calls_client_id(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    ClusteroneClient.start_job = mocker.Mock()

    CliRunner().invoke(cli, ['start', 'job', '--job-id', '1234-1234-1234'])

    ClusteroneClient.start_job.assert_called_with('1234-1234-1234')

def test_calls_client_name(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    ClusteroneClient.start_job = mocker.Mock()
    cmd.path_to_job_id = mocker.Mock(return_value="6789-6789-6789")


    CliRunner().invoke(cli, ['start', 'job', '--job-path', 'allgreed/my-tport-project/master'])

    cmd.path_to_job_id.assert_called_with('allgreed/my-tport-project/master', context=mocker.ANY)
    ClusteroneClient.start_job.assert_called_with('6789-6789-6789')



class ClusteroneException(Exception):
    def __init__(self):
        super(ClusteroneException, self).__init__()
        self.exit_code = 1


class InvalidProjectName(ClusteroneException):
    def __str__(self):
        return "Your project name is invalid, please be sure to only use letters, digits, hyphens, and underscores."


class InvalidDatasetName(ClusteroneException):
    def __str__(self):
        return "Your dataset name is invalid, please be sure to only use letters, digits, hyphens, and underscores."


class DuplicateProjectName(ClusteroneException):
    def __str__(self):
        return "Duplicate project name. Project names must be unique."


class DuplicateNotebookName(ClusteroneException):
    def __str__(self):
        return "Duplicate notebook name. Notebook names must be unique."


class UnsupportedSource(ClusteroneException):
    def __str__(self):
        return "Unsupported dataset source."

class BucketNameNotAvaliable(ClusteroneException):
    def __str__(self):
        return "The requested bucket name is not available. The bucket namespace is shared by all users of the system."

class DuplicateDatasetName(ClusteroneException):
    def __str__(self):
        return "Duplicate dataset name. Dataset names must be unique."


class LocalRepositoryFailure(ClusteroneException):
    def __str__(self):
        return "Couldn't find a Git repository. Please run this command in a valid repository or provide a valid repository path."


class RemoteAquisitionFailure(ClusteroneException):
    def __str__(self):
        return "Failed to get Git remote. Please contact Clusterone for support."


class LinkAlreadyPresent(ClusteroneException):
    def __str__(self):
        return "Cannot link. This repository already has a Clusterone remote."


class NonExistantProject(ClusteroneException):
    def __str__(self):
        return "Cannot resolve name to a project. Are you sure that such a project exists? If so, consider checking project owner and your permissions."


class NonExistantDataset(ClusteroneException):
    def __str__(self):
        return "Cannot resolve name to a dataset. Are you sure that such a dataset exists? If so, consider checking dataset owner and your permissions."


class JobNameConflict(ClusteroneException):
    def __init__(self, possible_ids):
        super(JobNameConflict, self).__init__()
        self.ids = possible_ids

    def __str__(self):
        return "{}{}{}".format(
            "Job name resolves to multiple IDs:\n",
            "".join(["{}\n".format(id) for id in self.ids]),
            "Cannot proceed. Please rerun this command with one of the IDs above.")


class _NonExistantJob(ClusteroneException):
    def __str__(self):
        return "There is no job with such ID. Are you sure that such a job exists?"


class NonExistantJob(ClusteroneException):
    def __str__(self):
        return "Cannot resolve name to a job. Are you sure that such a job exists?"


class _NonExistantNotebook(ClusteroneException):
    def __str__(self):
        return "There is no notebook with such ID. Are you sure that such a notebook exists?"


class NonExistantNotebook(ClusteroneException):
    def __str__(self):
        return "Cannot resolve name to a notebook. Are you sure that such a notebook exists?"


# TODO: Is this the most approperiate? See usage in tp_client
class LoginNotSupplied(ClusteroneException):
    def __str__(self):
        return "Action not authorized. Please log in or check your token."


class InternalServiceError(ClusteroneException):
    def __str__(self):
        return "Something went terribly wrong on our side, please contact us for immediate support."


class SoftInternalServiceError(ClusteroneException):
    """
    This is to be used in case of unknown problem if it may be resolved by logout-login action by the user.
    """

    def __str__(self):
        return "Something went wrong on our side. Could you please logout the login again? If the problem persits please contact us for immediate support."


class RunningJobRemoveAttempt(ClusteroneException):
    def __str__(self):
        return "This job is currently running."


class BusyProjectRemoveAttempt(ClusteroneException):
    def __str__(self):
        return "This project has currently running jobs."


class InsufficientResources(ClusteroneException):
    def __str__(self):
        return "You are exceeding you plan limit or do not have enough credits to proceed."


class BusyDatasetRemoveAttempt(ClusteroneException):
    def __str__(self):
        return "This dataset is currently in use by a running job. Removing it may cause the job to fail."


class NotSupported(ClusteroneException):
    def __str__(self):
        return "This CLI version is not up to date. Please run `pip install clusterone --upgrade`."


class LoginFailed(ClusteroneException):
    def __str__(self):
        return "Couldn't login, invalid username or password."


class WrongNumberOfGPUs(ClusteroneException):
    def __init__(self, instance_type, max_gpus):
        super(WrongNumberOfGPUs, self).__init__()

        self.instance_type = instance_type
        self.max_gpus = max_gpus

    def __str__(self):
        return "Wrong number of GPUs. '{}' instance's max number of GPUs is {}.\n" \
               "For more information please refer to the documentation: https://docs.clusterone.com/documentation/instance-types#instance-characteristics"\
            .format(self.instance_type, self.max_gpus)


class NoCommitsInProjectFound(ClusteroneException):
    def __str__(self):
        return 'No commits found in the project repository.'



from subprocess import call

import pytest

from clusterone.utilities import is_latest_version


@pytest.mark.slow()
def test_clusterone_throws_but_pytest_does_not():
    #TODO: Make it run in the approperiate directory regardless of invocation path
    bashCommand = "python clusterone/clusterone_cli.py"
    exit_code = call(bashCommand, shell=True)

    print("Please visit README-> Development -> Additional naming conventions and note about folder structure -> Caution. This may be a potential solution to your problem. Keep on Clusteronein'!")
    assert exit_code is 0

@pytest.mark.slow()
def test_clusterone_projects_dependency_met():
    """
    Clusterone project depend on 2 functions from the client.
    Thoose function should be properly exposed until further notice
    """

    try:
        from clusterone import get_data_path, get_logs_path
    except ImportError as exception:
        pytest.fail("Clusterone should propperly expose thoose 2 functions")


@pytest.mark.slow()
def test_version():
    """
    Makes sure that the version set in repo is >= the latest pypi version
    """

    assert is_latest_version()




# -*- coding: utf-8 -*-

import click
from clusterone import authenticate

@click.command()
@click.pass_obj
@authenticate()
def command(context):
    """
    Log out from Clusterone
    """

    context.session.delete()
    #TODO: Do a request to API to logout
    #TODO: Test this
    #TODO: Pylint tests
    #TODO: Refactor
    #TODO: Pylint



# -*- coding: utf-8 -*-


import os

import click
import coreapi
import py
from coreapi import Client
from coreapi import codecs
from coreapi import transports
from coreapi.auth import TokenAuthentication

from clusterone import client_exceptions
from clusterone.client_exceptions import WrongNumberOfGPUs, NoCommitsInProjectFound
from clusterone.utilities import lazy_property

# Python 2.7 compliance
try:
    from functools import lru_cache
except ImportError as exception:
    def identity(*args, **kwargs):
        """
        A dummy decorator that does nothing
        """

        def wrapper(function):
            """
            Like literally waste of space...
            """
            return function

        return wrapper


    lru_cache = identity

LOCAL_DEV = True if os.environ.get('JUST_DEBUG', 'False').upper() == "TRUE" else False


def get_data_path(dataset_name, local_root, local_repo, path):
    """
    Depending on the environment (clusterone cloud vs local) update path to data.
    Args:
        dataset_name: string, Clusterone dataset repository name, e.g. clusterone/mnist
        local_root: string, specifies the root directory for dataset.
              e.g. /home/username/datasets
        local_repo: string, specifies the repo name inside the root data path.
              e.g. mnist
        path: string, specifies the path inside the repository, (optional)
              e.g. train
    Returns:
        if code is running locally this function returns:
            local_root/local_repo/path
        if code runs on clusterone cloud this function returns:
            /data/dataset_name/path
    """
    #TODO: DO THIS PROPPERLY!!!!!!
    #environment = os.environ['CLUSTERONE_CLOUD']

    environment = os.environ.get('CLUSTERONE_CLOUD') or os.environ.get('TENSORPORT_CLOUD')
    environment = 'clusterone-cloud' if environment else "local"

    #TODO: Hotfixing TEN-819
    # Have no way of manually testing this
    # This should have a propper test case
    # Handling should be abstracted to a function - as it's the same here and in get_logs_path(root)
    # Need to show intent -> this is for platform indenpendence: converting Windows-style paths to UNIX-style paths. Both ways
    # Eg. /like/this/one <-> C:\like\this\one

    local_root = os.path.abspath(local_root)

    path = path.lstrip("/")  # preventing problems with user input

    if environment == 'local':
        return os.path.join(local_root, local_repo, path)
    elif environment == 'clusterone-cloud':
        return os.path.join('/data/', dataset_name, path)
    else:
        print("Environment is not valid. Assuming local.")
        return os.path.join(local_root, local_repo, path)


def get_logs_path(root):
    """
    Depending on the environment (clusterone cloud vs local) update path to logs.
    Args:
        root: string, specifies the directory for logs.
              e.g. /home/username/logs/mnist...
    Returns:
    """

    #TODO: See get_data_path() for more into, hotfixing TEN-819
    root = os.path.abspath(root)

    #TODO: DO THIS PROPPERLY!!!!!!
    #environment = os.environ['CLUSTERONE_CLOUD']

    environment = os.environ.get('CLUSTERONE_CLOUD') or os.environ.get('TENSORPORT_CLOUD')
    environment = 'clusterone-cloud' if environment else "local"

    if environment == 'local':
        return os.path.join(root)
    elif environment == 'clusterone-cloud':
        return os.path.join('/logs/')
    else:
        print("Environment is not valid. Assuming local.")
        return os.path.join(root)

class ClusteroneClient(object):
    """
    Clusterone API Client
    ^^^^^^^^^^^^^^^^^^^^^

    Clusterone Client (just for short) is a simple Python module that is used to get most up_to_date API schema,
    and execute API calls.

    """
    username = ""
    api_schema = {}
    use_in_memory_schema = True
    git_token = ""


    #TODO: Test this!
    @lazy_property
    def framework_slugs(self):
        #TODO: Move this to function signature after removing 2.7 compliance
        # type: () -> [str]

        return [framework['slug'] for framework in self.get_frameworks()]

    #TODO: Test this!
    @lazy_property
    def instance_types_slugs(self):
        #TODO: Move this to function signature after removing 2.7 compliance
        # type: () -> [str]

        return [machine['slug'] for machine in self.get_instance_types() if machine['show_for_workers']]


    #TODO: Test this!
    @lazy_property
    def ps_type_slugs(self):
        #TODO: Move this to function signature after removing 2.7 compliance
        # type: () -> [str]

        return [machine['slug'] for machine in self.get_instance_types() if machine['show_for_ps']]

    @property
    def matrix_url(self):
        #type: () -> str
        # TODO: After removeing Python 2.7 compliance move to function signautre
        #TODO: TEST THIS!!!

        return self.api_url.replace('api', 'matrix')

    def __init__(self, token="", username="", environment='', use_in_memory_schema=True, schema_json_path=None,
            api_url=None):

        if not api_url:
            # TODO: Do something nasty here
            pass

        self.api_url = api_url

        # TODO replace this with proper url generator with urllib ASAP
        self.api_schema_url = "{}{}".format(api_url, "schema/")
        self.environment = environment
        self.use_in_memory_schema = use_in_memory_schema

        self.token = token
        self.username = username

        # Load Schema from file or given path
        if not use_in_memory_schema:
            if schema_json_path:
                self.document_path = schema_json_path
            else:
                self.document_path = py.path.local(
                    click.get_app_dir('clusterone')).join('schema.json')

        # Init CoreAPI Client
        self.transport = self.get_transport()
        self.decoders = [codecs.CoreJSONCodec(), codecs.JSONCodec(), codecs.DownloadCodec()]
        self.api_client = Client(
            transports=[self.transport], decoders=self.decoders)

        # Get Schema
        self.api_schema = self.download_schema()

        super(ClusteroneClient, self).__init__()

    def get_transport(self):
        """
        Creates HTTPTransport with auth header information
        :param token:
        :return: HTTPTransport
        """

        # This is a dirty, dirty hack - disabling SSL checks!!!
        # Remove this after fixing SSL on Roche!
        import requests
        from functools import partial

        class partialmethod(partial):
            def __get__(self, instance, owner):
                if instance is None:
                    return self

                return partial(self.func, instance, *(self.args or ()), **(self.keywords or {}))

        # This disables the SSL warnings that are raised on insecure connection
        import warnings

        try:
            from urllib3.exceptions import InsecureRequestWarning
            warnings.simplefilter('ignore', InsecureRequestWarning)
        except ImportError:
            # on some weird configuration the Urllib3 is not present
            # this is better than worrying the users and breaking CLI
            warnings.simplefilter('ignore', Warning)


        if self.token:
            if LOCAL_DEV:
                print('Using Token: %s************' % self.token[:4])
            auth = TokenAuthentication(token=self.token, scheme='JWT')
            _session = requests.Session()
            _session.send = partialmethod(_session.send, verify=False)
            #TODO: Fix this header passing!!!!!!!!
            return transports.HTTPTransport(headers={'Authorization': '%s %s' % ('JWT', self.token)}, session=_session)
        else:
            _session = requests.Session()
            _session.send = partialmethod(_session.send, verify=False)
            return transports.HTTPTransport(session=_session)

    def get_schema_json(self):
        """
        Reads API SCHEMA from memory or JSON document stored in Clusterone local path
        :return:
        """
        if self.api_schema and self.use_in_memory_schema:
            return self.api_schema

        else:
            if not self.document_path.ensure():
                print('Cannot read from schema JSON, please relogin')
                return None
            store = self.document_path.open('rb')
            content = store.read()
            store.close()
            codec = codecs.CoreJSONCodec()
            return codec.load(content)

    def set_schema(self, document):
        """
        Creates Schema JSON in Clusterone local path
        :param document:
        :return:
        """
        codec = codecs.CoreJSONCodec()
        content = codec.dump(document)
        store = self.document_path.open('wb')
        store.write(content)
        store.close()
        print('Saved schema JSON')

    def download_schema(self):
        """
        Uses Basic Auth to get Schema JSON and save it
        :param document:
        :return:
        """
        API_SCHEMA = self.api_schema_url

        if LOCAL_DEV:
            print('Loading Clusterone API Schema: %s' % API_SCHEMA)
        try:
            api_schema = self.api_client.get(API_SCHEMA)

            if self.use_in_memory_schema:
                self.api_schema = api_schema
            else:
                self.set_schema(api_schema)

            return api_schema

        except coreapi.exceptions.ErrorMessage as e:
            # Reset token and try again
            # TODO this produces a nasty loop - will make hundrets of requests currently. Reset client triggers download schema...
            self.reset_client()
            api_schema = self.api_client.get(API_SCHEMA)

            if self.use_in_memory_schema:
                self.api_schema = api_schema
            else:
                self.set_schema(api_schema)

            return api_schema

        except Exception as e:
            print('Cannot download API Schema: %s' % str(e))
            raise

    def client_action(self, actions_keys, params=None, validate=True):
        """
        Wrapper for Client.action that set decoders, transports and schema document
        :param token:
        :param actions_keys:
        :param schema:
        :param params:
        :param validate:
        :return:
        """
        try:
            data = self.api_client.action(
                self.api_schema, actions_keys, params=params, validate=validate)
            if LOCAL_DEV:
                print("Success: %s" % data)
            return data
        except coreapi.exceptions.LinkLookupError as exc:
            raise client_exceptions.LoginNotSupplied()
        except coreapi.exceptions.ErrorMessage as exc:
            #TODO: Throtling, server error, gateway timeout, maintanance mode
            # Handle S3 Bucket name
            if hasattr(exc, 'error') and "S3 Bucket" in exc.error._data.get('messages', [''])[0]:
                raise client_exceptions.BucketNameNotAvaliable()
            if 'non_field_errors' in exc.error and exc.error['non_field_errors']:
                if (exc.error.title == '400 Bad Request') and exc.error._data['non_field_errors']._data[0] == 'Unable to log in with provided credentials.':
                    raise client_exceptions.LoginFailed()
            else:
                # TODO refactor the hell out of this
                if "400" in exc.error._title:
                    if "Enter a valid \"slug\" consisting of letters, numbers, underscores or hyphens." in exc.error._data.get("display_name", [''])[0]:
                        raise client_exceptions.InvalidProjectName()
                    elif "Project name is not unique" in exc.error._data.get("display_name", [''])[0]:
                        raise client_exceptions.DuplicateProjectName()
                    else:
                        raise
                if "403" in exc.error._title and 'You do not have enough resources to start' in exc.error._data['detail']:
                    raise client_exceptions.InsufficientResources()
                if "500" in exc.error._title:
                    raise client_exceptions.InternalServiceError()

                else:
                    if "404" not in exc.error._title:
                        print("API Error: %s" % exc.error)
                    else:
                        raise exc

    def reset_client(self):
        #TODO: Move this to function signature after removing 2.7 compliance
        # type: () -> None
        """
        Used to reset token, schema and set empty transrport method for client.
        Fixing CLUS-82.
        """

        self.token = None
        self.transport = self.get_transport()
        self.api_client = Client(
            transports=[self.transport], decoders=self.decoders)

        #TODO: Change to aquire schema in the future - to take advantage of cached schema and not making request
        self.api_schema = self.download_schema()

    # -----------------------------
    # Authentication
    def api_login(self, username, password):
            if LOCAL_DEV:
                print('Loading Initial Clusterone API Schema: %s' %
                      self.api_schema_url)

            # Login to get token
            response = self.client_action(actions_keys=['token', 'create'], params={
                'username': username,
                'password': password,
            }, validate=True)
            # Reset Token and transport headers
            if not response:
                message = "Couldn't login, please check your username and password"
                return None, message

            self.token = response.get('token')
            self.transport = self.get_transport()
            self.api_client = Client(
                transports=[self.transport], decoders=self.decoders)
            self.api_schema = self.download_schema()

            # Verify
            profile_response = self.client_action(actions_keys=['profile', 'list'], params={
            }, validate=True)

            self.git_token = profile_response.get('git_token')
            self.username = username

            return self.token

    def create_dataset(self, display_name, source='gitlab', description='Clusterone Dataset'):
        """
        Creates a dataset thru API
        :return: Dataset Unique Name
        """
        try:
            response = self.client_action(['datasets', 'owned', 'create'], params={
                'display_name': display_name,
                'description': description,
                'source': source,
                'parameters': {}
            }, validate=True)

            return response

        except client_exceptions.InvalidProjectName as exception:
            raise client_exceptions.InvalidDatasetName()
        except client_exceptions.DuplicateProjectName as exception:
            raise client_exceptions.DuplicateDatasetName()
        except client_exceptions.UnsupportedSource as exception:
            raise client_exceptions.UnsupportedSource()
        except client_exceptions.BucketNameNotAvaliable as exception:
            raise client_exceptions.BucketNameNotAvaliable()

    def create_project(self, display_name, description='Clusterone Project'):
        """
        Creates a project thru API
        :return: Project Unique Name
        """

        # Add Github support
        try:
            response = self.client_action(['projects', 'owned', 'create'], params={
                'display_name': display_name,
                'description': description,
                'parameters': {}
            }, validate=True)

            return response.get('name')

        #TODO: Somehow redo this -> catching all exceptions bloats sentry and causes further development problems
        #TODO: If this is the final Exception catcher then it shall: log to Sentry then kill the process
        except Exception as exception:
                if isinstance(exception, client_exceptions.InvalidProjectName):
                    raise client_exceptions.InvalidProjectName()
                elif isinstance(exception, client_exceptions.DuplicateProjectName):
                    raise client_exceptions.DuplicateProjectName()
                else:
                    print("Couldn't create a project thru API, error: %s" % str(exception))

                    return False

    def get_project(self, name, username=None):
        """
        Get Project from Repo
        :return: Project JSON
        """
        try:
            response = self.client_action(['projects', 'details', 'read'], params={
                'name': name,
                'username': username
            }, validate=False)
            return response
        except Exception as e:
            #TODO: This is crap, redo this ASAP
            if isinstance(e, client_exceptions.LoginNotSupplied):
                raise e
            if "404" in e.error._title:
                raise client_exceptions.NonExistantProject()
            else:
                print("Couldn't fetch a project, error: %s" % str(e))

                return False

    def get_projects(self, owner=None, writer=None, reader=None):
        """
        Get List of Owned Projects
        :return: List of Project JSONs
        """
        if owner:
            action = 'owned'
        elif writer:
            action = 'writable'
        elif reader:
            action = 'readable'
        else:
            action = 'readable'

        try:
            response = self.client_action(
                ['projects', action, 'list'], params={}, validate=False)
            return response
        except Exception as e:
            print("Couldn't fetch a project, error: %s" % str(e))

            return False

    def delete_project(self, name, username):
        """
        Deletes a Project
        :return:
        """
        try:
            response = self.client_action(['projects', 'details', 'delete'], params={
                'name': name,
                'username': username
            }, validate=True)
            return response
        except Exception as e:
            print("Couldn't delete a project, error: %s" % str(e))

            return False

    def delete_dataset(self, name, username):
        """
        Deletes a dataset
        :return:
        """
        # TODO: fix this
        try:
            response = self.client_action(['datasets', 'details', 'delete'], params={
                'name': name,
                'username': username
            }, validate=True)
            return response
        except Exception as e:
            print("Couldn't delete a dataset, error: %s" % str(e))

            return False

    def get_dataset(self, name, username):
        """
        Get Datasets from Repo
        :return: Project JSON
        """
        try:
            response = self.client_action(['datasets', 'details', 'read'], params={
                'name': name,
                'username': username
            }, validate=False)
            return response
        except Exception as e:
            if "404" in e.error._title:
                raise client_exceptions.NonExistantDataset()
            else:
                print("Couldn't fetch a dataset, error: %s" % str(e))

                return False

    def get_datasets(self, owner=None, reader=None, writer=None):
        """
        Get List of Owned Datasets
        :return: List of Project JSONs
        """
        # TODO: add various ownership levels
        try:
            if owner is not None:
                response = self.client_action(['datasets', 'owned', 'list'], params={
                }, validate=False)
            else:
                response = self.client_action(['datasets', 'list'], params={
                }, validate=False)
            return response
        except Exception as e:
            print("Couldn't fetch a datasets, error: %s" % str(e))

            return False

    def get_events(self):
        """
        Get List of user related events
        :return: List of Jobs JSONs
        """
        # TODO: add various ownership levels
        # TODO: what can this return? - discuss with Michał -> potencial edge cases
        # Error message: "Failed to query recent events. Perhaps you have not started a job yet. Try 'just run'."
        response = self.client_action(
            ['events', 'list'], params={}, validate=False)
        return response

    def create_event(self, job_id, event_type, event_level=20, event_content=None):
        """
        Create Event
        :return:
        """
        try:
            response = self.client_action(['events', 'create'], params={
                'job': job_id,
                'event_type': event_type,
                'event_level': event_level,
                'event_content': {}
            }, validate=True)
            return response
        except Exception as e:
            print("Couldn't create event, error: %s" % str(e))

            return False

    @lru_cache(maxsize=1)
    def get_instance_types(self, params=None):
        """
        Get List of Instance Types
        :return: List of AWS Instance Types JSONs
        """
        try:
            response = self.client_action(
                ['instance-types', 'list'], params=params, validate=False)
            return response.get('results')
        except Exception as e:

            print("Couldn't fetch list of instances, error: %s" % str(e))

    @lru_cache(maxsize=1)
    def get_frameworks(self, params=None):
        #TODO: What is a list with slug, name, number? A dict? A list of lists?
        """
        Get List of ML frameworks
        :return: List of frameworks with slug, name and number
        """
        try:
            response = self.client_action(
                ['frameworks', 'list'], params=params, validate=False)
            return response.get('results')
        #TODO: WHY LEO, WHY?!!!!!!!
        except Exception as e:

            print("Couldn't fetch list of frameworks, error: %s" % str(e))

    def get_tf_versions(self, params=None):
        """
        DEPRACATED - use get_frameworks instead
        :return: List of frameworks with slug, name and number
        """
        return self.get_frameworks(params)

    def get_jobs(self, params=None):
        """
        Get List of User Running Jobs
        :return: List of Jobs JSONs
        """
        try:
            if params:
                params.update({'limit': 200})
            else:
                params = {'limit': 200}
            response = self.client_action(
                ['jobs', 'list'], params=params, validate=False)
            return response
        except Exception as e:
            print("Couldn't fetch a project, error: %s" % str(e))

            return False

    def get_job(self, params=None):
        """
        Get a single Job
        :return: Job JSONs
        """
        try:
            response = self.client_action(
                ['jobs', 'read'], params=params, validate=False)
            return response
        except Exception as e:
            print("Couldn't fetch a job, error: %s" % str(e))

            return False

    def create_job(self, name, description=None, parameters=None):
        """
        create a Job, require repository ID
        :return:
        """
        request_parameters = self._prepare_request_parameters(name, description, parameters)

        try:
            response = self.client_action(['jobs', 'create'], params=request_parameters, validate=True)

            return response
        except Exception as e:
            try:
                error_message = e.error._data['parameters']['workers']['non_field_errors'][0]
                if "Wrong number of GPUs." in error_message:

                    words = error_message.split()
                    instance_type = words[4][1: -1]
                    max_gpus = words[11][:-1]

                    raise WrongNumberOfGPUs(instance_type, max_gpus)
            except (AttributeError, KeyError):
                pass
            try:
                if 'No commits found in the project repository.' in e.error._data['messages']:
                    raise NoCommitsInProjectFound()
            except (AttributeError, KeyError):
                pass

            print("Couldn't create a job, error: %s" % str(e))

            raise

    @classmethod
    def _prepare_request_parameters(cls, name, description, parameters):
        request_parameters = {}

        # if parameter not provided, it should not be sent to the API
        if parameters['datasets_set'] is not None:
            request_parameters['datasets_set'] = parameters['datasets_set']

        if parameters['git_commit_hash'] is not None:
            request_parameters['git_commit_hash'] = parameters['git_commit_hash']

        request_parameters['repository'] = parameters.pop('repository')
        request_parameters['display_name'] = name
        request_parameters['description'] = description
        request_parameters['parameters'] = parameters
        return request_parameters

    @staticmethod
    def _pop_from_source_to_target_if_exists(source, target, key):
        """If key exists in source (parameters) dict - move the key:value to target (request_parameters) dict"""
        try:
            value = source.pop(key)
            target[key] = value
        except KeyError:
            pass

    def delete_job(self, job_id):
        """
        Deletes a Job
        :return:
        """
        response = self.client_action(['jobs', 'delete'], params={
            'job_id': job_id,
        }, validate=True)
        return response

    def start_job(self, job_id):
        """
        Starts a Job
        :return:
        """
        try:
            response = self.client_action(['jobs', 'start'], params={
                'job_id': job_id,
            }, validate=True)
            return response
        except client_exceptions.InsufficientResources as exception:
            raise exception
        except Exception as exception:
            if "404" in exception.error._title:
                raise client_exceptions._NonExistantJob()

            print("Couldn't start a job, error: %s" % str(exception))

            return False

    def stop_job(self, job_id):
        """
        Stops a Job
        :return:
        """
        try:
            response = self.client_action(['jobs', 'stop'], params={
                'job_id': job_id,
            }, validate=True)
            return response
        except Exception as e:
            print("Couldn't stop a job, error: %s" % str(e))

            return False

    def get_file_list(self, job_id):
        """
        Get list of files/outputs
        :return:
        """
        try:
            response = self.client_action(['jobs', 'files', 'list'], params={
                'job_id': job_id,
            }, validate=True)
            if len(response) > 0:
                file_list = response[0]['contents']
                return file_list
            else:
                return []
        except Exception as e:
            print("Couldn't find a job or files, error: %s" % str(e))
            return False

    def download_file(self, job_id, filename):
        """
        Download file for Job
        :return:
        """
        try:

            download = self.client_action(['jobs', 'file', 'read'], params={
                'job_id': job_id,
                'filename': filename,
            }, validate=False)
            return download
        except Exception as e:
            print("Download file %s error: %s" % (filename, str(e)))

            return False



from uuid import UUID

from clusterone.utilities import path_to_job_id

def main(job_path_or_id, context=None):

    client = context.client

    job_path, job_id = None, None

    try:
        UUID(job_path_or_id)
        job_id = job_path_or_id
    except ValueError:
        job_path = job_path_or_id

    if job_path:
        job_id = path_to_job_id(job_path, context=context)

    job = client.get_job({"job_id": job_id})

    return job



import logging
from sys import exit


def run(scenario_cls):
    """
    Performs initializing, running and cleaning up after a scenario
    """

    scenario = scenario_cls()

    try:
        scenario.run()
        logging.info("SUCCESS!")
        exit(0)
    except AssertionError:
        logging.exception("TESTS FAILED AT:")
        exit(1)
    except Exception:
        logging.exception("TESTS FAILED DUE TO TEST CODE MALFUNCTION:")
        exit(2)
    finally:
        logging.debug("cleaning...")
        scenario.clean()
        logging.debug("cleaning done, exiting")



from .docstring import append_to_docstring

@append_to_docstring("dummy")
def dummy_function():
    """
    Sample docstring
    """

    pass

def test_docstring_appending(mocker):
    assert dummy_function.__doc__ == "\n    Sample docstring\n    dummy"



from collections import OrderedDict

GET_FRAMEWORKS_RESPONSE =[
        OrderedDict([('slug', 'pytorch-1.0.0'), ('name', 'pytorch'),
                    ('version', '1.0.0'), ('description', '')]),
        OrderedDict([('slug', 'tensorflow-1.4.0'), ('name', 'tensorflow'),
                    ('version', '1.4.0'), ('description', '')]),
        OrderedDict([('slug', 'tensorflow-1.5.0'), ('name', 'tensorflow'),
                    ('version', '1.5.0'), ('description', '')]),
        OrderedDict([('slug', 'tensorflow-1.3.0'), ('name', 'tensorflow'),
                    ('version', '1.3.0'), ('description', '')]),
        OrderedDict([('slug', 'tensorflow-1.2.0'), ('name', 'tensorflow'),
                    ('version', '1.2.0'), ('description', '')]),
        OrderedDict([('slug', 'tensorflow-1.0.0'), ('name', 'tensorflow'),
                    ('version', '1.0.0'), ('description', '')]),
  	]

GET_INSTANCE_TYPES_RESPONSE =[
        OrderedDict([
            ('slug', 'c4.2xlarge'),
            ('name', 'c4.2xlarge'),
            ('type', 'c4.2xlarge'),
            ('type_class', 'c'),
            ('cpu', 8),
            ('memory', 15),
            ('gpu', 0),
            ('description', ''),
            ('show_for_ps', True),
            ('show_for_workers', True),
            ('blessed', True),
            ('queue', 'tensorport-jobmaster-blessed'),
        ]),
        OrderedDict([
            ('slug', 'c4.2xlarge-spot'),
            ('name', 'c4.2xlarge Spot Instance'),
            ('type', 'c4.2xlarge'),
            ('type_class', 'c'),
            ('cpu', 8),
            ('memory', 61),
            ('gpu', 0),
            ('description', ''),
            ('show_for_ps', True),
            ('show_for_workers', True),
            ('blessed', False),
            ('queue', 'tensorport-jobmaster'),
        ]),
        OrderedDict([
            ('slug', 'p2.xlarge'),
            ('name', 'p2.xlarge'),
            ('type', 'p2.xlarge'),
            ('type_class', 'c'),
            ('cpu', 4),
            ('memory', 61),
            ('gpu', 1),
            ('description', ''),
            ('show_for_ps', False),
            ('show_for_workers', True),
            ('blessed', True),
            ('queue', 'tensorport-jobmaster-blessed'),
        ]),
        OrderedDict([
            ('slug', 'p3.2xlarge'),
            ('name', 'p3.2xlarge'),
            ('type', 'p3.2xlarge'),
            ('type_class', 'p'),
            ('cpu', 8),
            ('memory', 61),
            ('gpu', 1),
            ('description', ''),
            ('show_for_ps', False),
            ('show_for_workers', True),
            ('blessed', True),
            ('queue', 'tensorport-jobmaster-blessed'),
        ]),
        OrderedDict([
            ('slug', 't2.small'),
            ('name', 't2.small'),
            ('type', 't2.small'),
            ('type_class', 'c'),
            ('cpu', 1),
            ('memory', 2),
            ('gpu', 0),
            ('description', ''),
            ('show_for_ps', True),
            ('show_for_workers', True),
            ('blessed', True),
            ('queue', 'tensorport-jobmaster-blessed'),
        ]),
        OrderedDict([
            ('slug', 't2.small-spot'),
            ('name', 't2.small Spot Instance'),
            ('type', 't2.small'),
            ('type_class', 'c'),
            ('cpu', 1),
            ('memory', 2),
            ('gpu', 0),
            ('description', ''),
            ('show_for_ps', True),
            ('show_for_workers', True),
            ('blessed', False),
            ('queue', 'tensorport-jobmaster'),
        ]),
    ]



from collections import OrderedDict

from click.testing import CliRunner

from clusterone import ClusteroneClient
from clusterone.clusterone_cli import cli
from clusterone.commands.get.dataset import cmd
from clusterone.mocks.dataset import GET_DATASET_TEST


def test_passing_project_path(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    cmd.path_to_dataset = mocker.Mock()

    CliRunner().invoke(cli, ['get', 'dataset', 'some_dataset_name'])

    cmd.path_to_dataset.assert_called_with('some_dataset_name', context=mocker.ANY)

def test_table_call(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    cmd.path_to_dataset = mocker.Mock(return_value=GET_DATASET_TEST)
    cmd.make_table = mocker.Mock(return_value="Example table string")

    CliRunner().invoke(cli, ['get', 'dataset', 'some_dataset_name'])

    cmd.make_table.assert_called_with([
        ('Name', 'test_kwiatek'),
        ('Id', 'ab3d71a8-6c31-40fc-a761-d1b5f7b32729'),
        ('Source', 'gitlab'),
        ('URL', 'https://git.tensorport.com/smeler/test_kwiatek.git'),
        ], header=['Property', 'Value'])

def test_printing_out_table(mocker):
    mocker.patch.object(ClusteroneClient, '__init__', autospec=True, return_value=None)
    cmd.path_to_dataset = mocker.Mock(return_value=GET_DATASET_TEST)
    cmd.make_table = mocker.Mock(return_value="Example table string")

    result = CliRunner().invoke(cli, ['get', 'dataset', 'some_dataset_name'])

    assert result.output == "Example table string\n"



"""
Building blocks of a functional tests

Contains high level action as understood by the business.
"""

import subprocess
import logging

from .utilities import just_exec, call, poll
from .types import Output
from coreapi.exceptions import ErrorMessage


def set_endpoint(url):
    just_exec("config endpoint {}".format(url))

    logging.info("using endpoint {}".format(url))


def register_user(user_credentials, client):

    response = client.client_action(['register', 'create'], params={
        'username': user_credentials.username,
        'email': user_credentials.email,
        'password': user_credentials.password,
    }, validate=True)

    assert response['username'] == user_credentials.username, "Registration failure"
    assert response['email'] == user_credentials.email, "Registration failure"

    logging.info("registration complete")


def login(user_credentials):

    output = just_exec(
        "login --username {} --password {}".format(user_credentials.username, user_credentials.password))

    assert "successful" in output, "Login failure"

    logging.info("login completed")


def logout():
    just_exec("logout")

    logging.debug("credentials clear")


def init_gitlab_project_from_clusterone_samples(name, project_url):
    call("git clone -q {}".format(project_url))
    folder_name = project_url.split("/")[-1]
    project_initialization_output = just_exec("init project {} -r ./{}".format(name, folder_name))
    subprocess.call("cd {}; git push -q clusterone".format(folder_name), shell=True)

    assert project_initialization_output.split('\n')[1].endswith('{}.git'.format(name)), "Project creation failure"

    logging.info("local and remote project clone completed")

    # check if commits appeared
    poll(just_exec, args=tuple(["create job single --project {} --name {}".format(name, "test_project_has_commits")]), timeout=60, interval=10)


def create_job(project_name, job_name, module="main", framework=None, dataset=None):

    job_state = just_exec("get jobs")
    assert job_name not in job_state, "???"

    framework = ("--framework {}".format(framework)) if framework else ""
    dataset = "--datasets {}".format(dataset) if dataset else ""

    just_exec("create job distributed --name {} --project {} --module {} --time-limit 1h {} {}".format(job_name, project_name, module, framework, dataset))

    job_state = just_exec("get jobs")
    assert job_name in job_state, "Job creation failure"

    logging.info("job successfully created")


def start_job(project_name, job_name):
    just_exec("start job -p {}/{}".format(project_name, job_name))
    job_status = just_exec("get job {}/{}".format(project_name, job_name))

    assert "starting" in job_status, "Job start failure"

    logging.info("job successfully started")


def check_outputs(desired_outputs, job_name, client=None):

    def download_file(filename):
        response = client.client_action(['jobs', 'file', 'read'], params={
            'job_id': job_id,
            'filename': filename
        }, validate=True)

        return response

    job_data = just_exec("get jobs")
    job_data_parsed = job_data.split()
    job_id = job_data_parsed[job_data_parsed.index("None/{}".format(job_name)) + 2]

    try:
        response = client.client_action(['jobs', 'files', 'list'], params={
            'job_id': job_id,
        }, validate=True)
    except ErrorMessage as exception:

        if "404 Not Found" in str(exception.error):
            outputs = []
        else:
            raise
    else:
        outputs = response[0]['contents']

    if desired_outputs == Output.ANY:
        assert outputs, "ANY output acquisition failure"
    elif desired_outputs == Output.NONE:
        assert outputs == []
    elif desired_outputs == Output.CHECKPOINT:
        assert any({"ckpt" in output_file['name'] for output_file in outputs}), "CHECKPOINT output acquisition failure"
    elif desired_outputs == Output.LOSS:
        for worker_log in {output_file['name'] for output_file in outputs if "worker" in output_file['name'] and output_file['name'].endswith(".txt")}:
            assert "loss" in download_file(worker_log).lower(), "LOSS output acquisition failure"
    else:
        raise ValueError("Desired_outputs parameter must be correct enum value")

    logging.info("outputs match target {}".format(desired_outputs))



from collections import OrderedDict

GET_DATASET_TEST = OrderedDict(
    [('id', 'ab3d71a8-6c31-40fc-a761-d1b5f7b32729'), ('display_name', 'test_kwiatek'), ('icon', 'fa-btc'),
     ('color', '#47c5d1'), ('name', 'test_kwiatek'), ('source', 'gitlab'), ('full_name', 'smeler/test_kwiatek'),
     ('dataset', 'ab3d71a8-6c31-40fc-a761-d1b5f7b32729'), ('parameters', OrderedDict()), ('description', ''),
     ('http_url_to_repo', 'https://git.tensorport.com/smeler/test_kwiatek.git'), ('git_auth_link',
                                                                                  'https://smeler:28ea32b920da06b2350531cd6d26de17ab231906@git.tensorport.com/smeler/test_kwiatek.git'),
     ('parent_repository', None), ('parent_repository_name', None), ('parent_repository_owner', None),
     ('created_by', 269), ('members', [OrderedDict([('id', 269), ('access_level', 'owner'), ('username', 'smeler'), (
    'photo_url', 'https://tensorport-api-development.s3.amazonaws.com/images/empty_image_400_400.png'),
                                                    ('email', 'smeler@gormanet.pl'), ('first_name', ''),
                                                    ('last_name', ''), ('profile_url', 'QA3Y0F86M0XW'),
                                                    ('date_joined', '2017-07-24T15:15:49Z'),
                                                    ('last_login', '2018-04-05T08:06:13.447233Z')])]), ('commits', []),
     ('files', None), ('owner', OrderedDict([('id', 269), ('access_level', 'owner'), ('username', 'smeler'), (
    'photo_url', 'https://tensorport-api-development.s3.amazonaws.com/images/empty_image_400_400.png'),
                                             ('email', 'smeler@gormanet.pl'), ('first_name', ''), ('last_name', ''),
                                             ('profile_url', 'QA3Y0F86M0XW'), ('date_joined', '2017-07-24T15:15:49Z'),
                                             ('last_login', '2018-04-05T08:06:13.447233Z')])), ('is_public', False),
     ('created_at', '2018-03-29T12:37:28.907855Z'), ('modified_at', '2018-04-11T13:47:25.121852Z'), ('tags', []),
     ('size', 0)])



try:
    from math import inf
except ImportError as exception:
    inf = float('inf')

import click
from click import IntRange
from click.exceptions import BadParameter

from clusterone import client
from clusterone.utilities import Choice
from .base_cmd import job_base_options, base

PS_REPLICAS_WARNING_THRESHOLD = 5
WORKER_REPLICAS_WARNING_THRESHOLD = 10

# TODO: Move this to utilities
POSITIVE_INTEGER = IntRange(1, inf)


@job_base_options()
@click.option(
    '--worker-type',
    type=Choice(client.instance_types_slugs),
    default="t2.small",
    help="Type of worker instances")
@click.option(
    '--ps-type',
    type=Choice(client.ps_type_slugs),
    default="t2.small",
    help="Type of parameter server instances")
@click.option(
    '--worker-replicas',
    type=POSITIVE_INTEGER,
    default=2,
    help="Number of worker instances",
)
@click.option(
    '--ps-replicas',
    type=POSITIVE_INTEGER,
    default=1,
    help="Number of parameter server instances",
)
def command(context, custom_arguments, **kwargs):
    """
    Creates a distributed job.
    """

    client = context.client
    job_configuration = base(context, kwargs, module_arguments=custom_arguments)

    job_configuration['parameters']['mode'] = "distributed"

    ps_replicas, worker_replicas = kwargs['ps_replicas'], kwargs['worker_replicas']

    if ps_replicas > PS_REPLICAS_WARNING_THRESHOLD:
        click.echo("Caution: You are creating a job with more than {} parameter servers.".format(PS_REPLICAS_WARNING_THRESHOLD))
    if worker_replicas > WORKER_REPLICAS_WARNING_THRESHOLD:
        click.echo("Caution: You are creating a job with more than {} workers.".format(WORKER_REPLICAS_WARNING_THRESHOLD))

    # TODO: Levearge dynamic frameworks here - requires API support - framework should have a list of supported modes
    # TODO: Test this
    if job_configuration['parameters']['framework']['slug'] == 'pytorch-1.0.0':
        raise BadParameter("PyTorch is only supported for single jobs. For PyTorch please use `create job single [...]` instead.", param_hint="--framework")

    job_configuration['parameters']['workers'] = \
        {
            'slug': kwargs['worker_type'],
            'replicas': worker_replicas,
        }

    gpu_count = kwargs.get('gpu_count')
    if gpu_count:
        job_configuration['parameters']['workers']['gpu_count'] = gpu_count

    job_configuration['parameters']['parameter_servers'] = \
        {
            'slug': kwargs['ps_type'],
            'replicas': ps_replicas,
        }

    client.create_job(
        job_configuration['meta']['name'],
        description=job_configuration['meta']['description'],
        parameters=job_configuration['parameters'],
    )



from click import ParamType
from . import main


def test_subclassing():
    assert issubclass(main.Notebook, ParamType)


def test_notebook_acquisition(mocker):
    main.JustNotebook = mocker.Mock()

    main.Notebook.convert(main.Notebook.__new__(main.Notebook), "elorap", "", "")

    main.JustNotebook.from_clusterone.assert_called_with(mocker.ANY, "elorap")



from collections import OrderedDict

import pytest
from pytest import raises
from click.exceptions import BadParameter

from clusterone.persistance.session import Session
from clusterone.clusterone_cli import Context
from clusterone.commands.create.job import base_cmd


def test_python_version_aliases(mocker):
    base_cmd.client = mocker.Mock()
    base_cmd.path_to_project = mocker.Mock(return_value={'id': "project-id-123456", "commits": [OrderedDict([('id', '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1')]), OrderedDict([('id', '4a82d16c79elorapc7973af38f2f5ba4eac0cd2d1')])]})
    base_cmd.time_limit_to_minutes = mocker.Mock(return_value=123456)
    session = Session()
    session.load()
    context = Context(None, session, None)

    result = base_cmd.base(context, {'framework': 'tensorflow-130', 'project_path': 'mnist-demo', 'python_version': '3', 'requirements': None, 'commit': 'latest', 'name': 'old-morning-562', 'module': 'mymodule.py', 'package_manager': 'pip', 'package_path': '', 'description': '', 'time_limit': 2880, 'datasets': '',}, module_arguments=())


    assert result['parameters']['python_version'] == '3.6'

    result = base_cmd.base(context, {'framework': 'tensorflow-130', 'project_path': 'mnist-demo', 'python_version': '2', 'requirements': None, 'commit': 'latest', 'name': 'old-morning-562', 'module': 'mymodule.py', 'package_manager': 'pip', 'package_path': '', 'description': '', 'time_limit': 2880, 'datasets': '',}, module_arguments=())


    assert result['parameters']['python_version'] == '2.7'

def test_package_manager_aliases(mocker):
    base_cmd.client = mocker.Mock()
    base_cmd.path_to_project = mocker.Mock(return_value={'id': "project-id-123456", "commits": [OrderedDict([('id', '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1')]), OrderedDict([('id', '4a82d16c79elorapc7973af38f2f5ba4eac0cd2d1')])]})
    base_cmd.time_limit_to_minutes = mocker.Mock(return_value=123456)
    session = Session()
    session.load()
    context = Context(None, session, None)

    result = base_cmd.base(context, {'framework': 'tensorflow-130', 'project_path': 'mnist-demo', 'python_version': '3', 'requirements': None, 'commit': 'latest', 'name': 'old-morning-562', 'module': 'mymodule.py', 'package_manager': 'anaconda', 'package_path': '', 'description': '', 'time_limit': 2880, 'datasets': '',}, module_arguments=())


    assert result['parameters']['package_manager'] == 'conda'

def test_default_requirement_conda(mocker):
    base_cmd.client = mocker.Mock()
    base_cmd.path_to_project = mocker.Mock(return_value={'id': "project-id-123456", "commits": [OrderedDict([('id', '4a82d16c7995856c7973af38f2f5ba4eac0cd2d1')]), OrderedDict([('id', '4a82d16c79elorapc7973af38f2f5ba4eac0cd2d1')])]})
    base_cmd.time_limit_to_minutes = mocker.Mock(return_value=123456)
    session = Session()
    session.load()
    context = Context(None, session, None)

    result = base_cmd.base(context, {'framework': 'tensorflow-1.3.0', 'project_path': 'mnist-demo', 'python_version': '3', 'requirements': None, 'commit': 'latest', 'name': 'old-morning-562', 'module': 'mymodule.py', 'package_manager': 'anaconda', 'package_path': '', 'description': '', 'time_limit': 2880, 'datasets': '',}, module_arguments=())


    assert result['parameters']['requirements'] == 'requirements.yml'


def test_custom_user_arguments(mocker):
    assert base_cmd.validate_module_arguments(mocker.ANY, mocker.ANY, ('--custom-dummy-argument0', 'dummy0', '--custom-dummy-argument1', 'dummy1')) == {'custom-dummy-argument0': 'dummy0', 'custom-dummy-argument1': 'dummy1'}

def test_custom_user_arguemnts_missing_value(mocker):
    with raises(BadParameter):
        base_cmd.validate_module_arguments(mocker.ANY, mocker.ANY, ('--custom-dummy-argument0', '--custom-dummy-argument1', 'dummy1'))
